#include "engine.h"
#include "render.h"



/*
 Transform vertex World->Camera.
*/
void XformVertex(Camera *C,float *src,float *dest)
{
	float t[3];

	FVec_SubFVec(src,C->c,t);
	FMatrix_InvMulFVec(C->ToWorld,t,dest);
}

/*
 Project vertex in camera space.
*/
void ProjectVertex(Camera *C,float *src,float *dest)
{
	float proj_x,proj_y, inv_z;

	inv_z=1.0f/src[2];
	proj_x=C->proj*inv_z;
	proj_y=proj_x*C->proj_ratio;
	dest[0]=C->mid_x+src[0]*proj_x;
	dest[1]=C->mid_y+src[1]*proj_y;
	dest[2]=inv_z;
}

/* 
 Initialize camera
*/
void Camera_Init(Camera *C, float fov,int w,int h,int x,int y)
{
	float s,c,temp[3]={0,0,0};

	C->width=w; C->height=h;
	C->sx=x; C->sy=y;

	C->mid_x=(float)(w>>1)-0.5f;
	C->mid_y=(float)(h>>1)-0.5f;

	s=(float)sin(fov*0.5f);
	c=(float)cos(fov*0.5f);

// Make projection constants
	C->proj=(c*(float)(w>>1))/s;

// Projection ratio (1.0 for square pixels)
//	C->proj_ratio=((float)Render_GetScrWidth())*0.75;
//	C->proj_ratio=((float)Render_GetScrHeight())/C->proj_ratio;
	C->proj_ratio=640.0f*0.75f;
	C->proj_ratio=480.0f/C->proj_ratio;

// Init coords and rotate
	FVec_Copy(C->c,temp);
	FMatrix_Identity(C->ToWorld);
	C->fov=fov;
}

/*
 Move camera in world space
*/
void Camera_Move(Camera *C, float *delta)
{
	float t[3];

	FMatrix_MulFVec(C->ToWorld,delta,t);
	FVec_AddFVec(C->c,t,C->c);
}

void Camera_MakeXform(Camera *C)
{
	FMatrix_Identity(C->ToWorld);
	FMatrix_PreZRot(C->ToWorld,C->yaw);
	FMatrix_PreYRot(C->ToWorld,C->roll);
	FMatrix_PreXRot(C->ToWorld,C->pitch);
}

/*
 Rotate camera in world space
*/
/*void Camera_Rotate(Camera *C, float *delta)
{
	FMatrix_PreYRot(C->ToWorld,delta[1]);
	FMatrix_PreXRot(C->ToWorld,delta[0]);
	FMatrix_PreZRot(C->ToWorld,delta[2]);
}
*/
