#include "engine.h"
#include "render.h"
#include "h_sound.h"
#include "h_input.h"

#include "progs\progs.h"	// Just a hack for extcamera


// The renderer
RenderExport *render;
static HMODULE renderModule;
static RenderExport *(*GetRender)(RenderImport *rimp);

// For connection to DI
HWND clHwnd;
HINSTANCE clInstance;

static int Counter=-1, CounterStartTicks, CounterStartFrames;

static Camera MainCamera;

int do_count=0, do_light=0, do_atmo=0;

// From PAL.C
extern byte   RGBPalette[768];

extern void Overlay_Render(Camera *C);
extern void Overlay_Attack(void);


RenderImport rimp;


int Client_Init(EngineImport *eimp)
{
	Camera *C=&MainCamera;


// Load rendering DLL
	renderModule=LoadLibrary("rogl");

	if (!renderModule) return FALSE;

// Map the function
	GetRender = (void *)GetProcAddress( renderModule, "GetRender");

	if (!GetRender) {
		FreeLibrary(renderModule);
		return FALSE;
	}

// Setup import data for the DLL
	rimp.hInstance=eimp->hInstance;
	rimp.Palette=RGBPalette;
	rimp.Paused=&Paused;
	rimp.Done=&Done;
	rimp.LightMode=&do_light;
	rimp.DebugMode=&do_count;

// Setup the renderer DLL
	render = GetRender(&rimp);



// Load the palette
	InitPalette("POLY.PAL");

// Init the renderer
	clInstance=eimp->hInstance;

	clHwnd=render->init(
			eimp->renderWidth,
			eimp->renderHeight,
			eimp->renderFlags);


// Init the camera
	Camera_Init(C,PI/2.0,render->getScrWidth(),render->getScrHeight(),0,0);

// Init the input system
	Input_Init(INPUT_MOUSE|INPUT_VR|INPUT_KEYBOARD);

// Init sound system
	SND_Init();

	return TRUE;
}


void Client_Shut() 
{
	FreeWorld();
	Client_Disconnect();
	Input_Shut();
	SND_Shut();
	FreePalette();

	render->shut();
	FreeLibrary(renderModule);
}


void SyncCamera(Camera *C,ClientFrame *cl)
{
	char str[64];


	FVec_Copy(cl->cam_pos,C->c);
	C->c[2]+=16+cl->cam_height;

	C->pitch=cl->head_pitch-PI/2;
	C->roll=cl->body_roll+cl->head_roll;
	C->yaw=cl->cam_angle+cl->head_yaw;

	if (Counter>=0) {
		if (Counter==128) {
			Counter=-1;
			sprintf(str,"Timer: %d %f\n",TotalTicks-CounterStartTicks,
				(100.0*(float)(TotalFrames-CounterStartFrames))/
				((float)(TotalTicks-CounterStartTicks)));
			MessageBox(NULL,str,"info",MB_OK);
		
		}
		else {
			C->yaw=(PI*(float)Counter)/64.0f;
			C->pitch=-PI/2;
			C->roll=0;
			Counter++;
		}
	}

	Camera_MakeXform(C);
}


void Client_ProcessFrame(ClientFrame *cl)
{
	Camera *C=&MainCamera;
	POB *pob;
	float dir[3], up[3];
	int i, type;


// Update camera position
//	if (DoCameraSync)
		SyncCamera(C,cl);

// No world bsp!
	if (!cl->MainBSP) return;

// Set SND listener position to Camera
	for(i=0;i<3;i++) {
		dir[i]=C->ToWorld[i][1];
		up[i]=C->ToWorld[i][2];
	}

	SND_SetListener(C->c,dir,up);

// Pre-process all visible objects
	pob=cl->VisObjects;
	for(i=0;i<cl->NumVisObjects;i++,pob++) {
		Prog_Process(pob);
	}

// Update particles
	UpdateParticles();

// Set Renderer camera
	render->beginFrame(C);

// Render bsp world
	render->beginBSP(cl->MainBSP);

// Render all visible bsp models
	pob=cl->VisObjects;
	for(i=0;i<cl->NumVisObjects;i++,pob++) {
		if (Prog_GetProgFlags(pob->prog_idx)&PF_TYPE_BSP_MODEL) {
			Prog_Render(pob);
		}
	}
	render->endBSP();

// Render all visible objects
	pob=cl->VisObjects;
	for(i=0;i<cl->NumVisObjects;i++,pob++) {
		type=Prog_GetProgFlags(pob->prog_idx);
		if (type&(PF_TYPE_MODEL|PF_TYPE_SPRITE))
			Prog_Render(pob);
		pob->flags=0;
	}

// Render particles
	render->drawParticles(firstActive,PARTICLE_DOT);

// Render overlay(s)
	if (cl->player_flags&(PLAYER_FLAG_START_ATTACK))
		Overlay_Attack();
	if ((cl->player_flags&(PLAYER_FLAG_DEAD|PLAYER_FLAG_EXTCAMERA))==0)
		Overlay_Render(C);


// Finish off rendering the frame
	render->endFrame();

}


/******************************************************************
				Process the input
*******************************************************************/

static char levels[][16]={
	"jail1.bsp","jail2.bsp","jail3.bsp", "base1.bsp",
};
static int next_level=0;

/*
  Process keyboard input for the player object (kbd handler)
*/
void Game_KbdInput(ClientFrame *cl)
{
	int kTAB_pressed=0, kENTER_pressed=0;
	char params[32];

	if (KeyPressed==kESC) {
		KeyPressed=0;
		Done=1;
	}

	if (Keys[kTAB]&&!kTAB_pressed) {
		kTAB_pressed=1;

//		DoCameraSync^=1;
//		do_save^=1;
//		SavePos("SAVEPOS.POS",&MainCamera);

		if (NumPlayers>0) {
			Server_NewWorld(levels[next_level]);
			next_level++;
			next_level%=4;
		}

		KeyPressed=0;
	}
	else
		kTAB_pressed=0;

	if (Keys[kENTER]&&!kENTER_pressed) {
		kENTER_pressed=1;
		KeyPressed=0;
		if (NumPlayers>0) {
			if (Players[0].camobj_idx<0) {
				sprintf(params,"\"follow\" \"%d\"",Players[0].obj_idx);
				Players[0].camobj_idx=World_AddObject(PROG_EXTCAMERA,params);
			}
			else {
				World_RemoveObject(Players[0].camobj_idx);
				Players[0].camobj_idx=-1;
			}
		}
	}
	else
		kENTER_pressed=0;

// Lamer kbd setup ;)
	if (Keys[kUARROW])
		cl->player_action|=ACTION_FORWARD;
	if (Keys[kDARROW])
		cl->player_action|=ACTION_BACKWARD;
	if (Keys[kLARROW]) {
		if (Keys[kRIGHTALT]||Keys[kLEFTALT])
			cl->player_action|=ACTION_MOVE_LEFT;
		else
			cl->player_action|=ACTION_TURN_LEFT;
	}
	if (Keys[kRARROW]) {
		if (Keys[kRIGHTALT]||Keys[kLEFTALT])
			cl->player_action|=ACTION_MOVE_RIGHT;
		else
			cl->player_action|=ACTION_TURN_RIGHT;
	}

// Mouser kbd setup
	if (Keys[kE])
		cl->player_action|=ACTION_FORWARD;
	if (Keys[kD])
		cl->player_action|=ACTION_BACKWARD;
	if (Keys[kS])
		cl->player_action|=ACTION_MOVE_LEFT;
	if (Keys[kF])
		cl->player_action|=ACTION_MOVE_RIGHT;

// Action keys
	if (Keys[kRIGHTCTRL]||Keys[kLEFTCTRL])
		cl->player_action|=ACTION_ATTACK;

	if (Keys[kSPACE])
		cl->player_action|=ACTION_JUMP;

// !!! HACKS !!!
	if (Keys[kA]) 
		do_atmo^=1;
	if (Keys[kZ]) 
		Objects[Players[0].obj_idx].pob.c[2]+=4.0;

	if (Keys[kL]) 
		do_light=(do_light+1)%3;

	if (Keys[kC]) 
		do_count=1;

	if (Keys[kO]&&Counter<0) {
		Counter=0;
		CounterStartTicks=TotalTicks;
		CounterStartFrames=TotalFrames;
	}
}

/*
  Get and process mouse input
*/
void Game_MouInput(ClientFrame *cl)
{

	if (MouPress&1)
		cl->player_action|=ACTION_ATTACK;

	if (MouPress&2) {
		cl->player_action|=ACTION_JUMP;
	}

	if (MouDeltaY) {
		cl->head_pitch+=(float)MouDeltaY*PI/512;
		if (cl->head_pitch>PI/2)
			cl->head_pitch=PI/2;
		else if (cl->head_pitch<-PI/2)
			cl->head_pitch=-PI/2;
	}

	if (MouDeltaX) {
		if ((cl->player_flags&PLAYER_FLAG_DEAD)==0) {
			cl->body_yaw+=(float)-MouDeltaX*PI/512;
		}
	}
}


/*
  Process VR input for a player
*/
void Game_VRInput(ClientFrame *cl)
{
	cl->head_pitch=(((float)VR_Pitch)/32768.0f)*PI;
	cl->head_yaw=  (((float)VR_Yaw  )/32768.0f)*PI;
	cl->head_roll= (((float)VR_Roll )/32768.0f)*PI;
}



void Client_GetInput(ClientFrame *cl)
{
	int flags;

	flags=Input_GetData();

	if (flags&INPUT_MOUSE) Game_MouInput(cl);
	if (flags&INPUT_VR) Game_VRInput(cl);
	if (flags&INPUT_KEYBOARD) Game_KbdInput(cl);
}

