
#include <windows.h>
#include <dinput.h>


// DirectInput stuff (must be called after DD_Init)
BOOL    DI_Init(void);
void    DI_End(void);
void    DI_GetMouseData(void);
void    DI_GetKeyboardData(void);



extern "C" {

	#include "h_input.h"

// Info from the client
	extern HWND clHwnd;
	extern HINSTANCE clInstance;

// From the main engine
	extern int Paused;

// Mouse input
	int MouX, MouY, MouDeltaX, MouDeltaY, MouPress, MouRelease;

// Kbd input
	byte Keys[256], KeyPressed;

// VR stuff (placeholder)
	int VR_Yaw=0, VR_Pitch=0, VR_Roll=0;


	int Input_Init(int flags) {
		return (DI_Init());
	}

	void Input_Shut(void) {
		DI_End();
	}

	int Input_GetData(void) {
		DI_GetMouseData();
		DI_GetKeyboardData();
		return(INPUT_KEYBOARD|INPUT_MOUSE);
	}

// Special func, called from engine msg processing
	void    DI_SyncAcquire(void);

}




// DirectInput stuff

LPDIRECTINPUT			pdi = NULL;
LPDIRECTINPUTDEVICE		pMouse = NULL;
LPDIRECTINPUTDEVICE		pKeyboard = NULL;




// Init all DI devices (mouse, kbd). Must be called after either 
// client has initialized a renderer

BOOL DI_Init(void)
{
	HRESULT hr;

// Already initialized
	if (pdi) return TRUE;

// Init DirectInput
	hr = DirectInputCreate(clInstance,DIRECTINPUT_VERSION,&pdi,NULL);
	if (FAILED(hr)) return FALSE;

// Init mouse
	hr = pdi->CreateDevice(GUID_SysMouse,&pMouse,NULL);
	if (FAILED(hr)) return FALSE;

	hr = pMouse->SetDataFormat(&c_dfDIMouse);
	if (FAILED(hr)) return FALSE;

	hr = pMouse->SetCooperativeLevel(clHwnd,DISCL_EXCLUSIVE|DISCL_FOREGROUND);
	if (FAILED(hr)) return FALSE;

// Init keyboard
	hr = pdi->CreateDevice(GUID_SysKeyboard,&pKeyboard,NULL);
	if (FAILED(hr)) return FALSE;

	hr = pKeyboard->SetDataFormat(&c_dfDIKeyboard);
	if (FAILED(hr)) return FALSE;

	hr = pKeyboard->SetCooperativeLevel(clHwnd,DISCL_NONEXCLUSIVE|DISCL_FOREGROUND);
	if (FAILED(hr)) return FALSE;

// Since we already have a window, acquire mouse and kbd
	DI_SyncAcquire();

	return TRUE;
}


void DI_End(void)
{

// Shut keyboard
	if (pKeyboard) {
		pKeyboard->Unacquire();
		pKeyboard->Release();
		pKeyboard = NULL;
	}

// Shut mouse
	if (pMouse) {
		pMouse->Unacquire();
		pMouse->Release();
		pMouse = NULL;
	}

// Shut DirectInput
	if (pdi) {
		pdi->Release();
		pdi = NULL;
	}
}


void DI_GetMouseData(void)
{
	DIMOUSESTATE dims;
	HRESULT hr;

	if (!pMouse) return;

again:;
	hr = pMouse->GetDeviceState(sizeof(DIMOUSESTATE), &dims);
	if (hr == DIERR_INPUTLOST) {
		hr = pMouse->Acquire();
			if (SUCCEEDED(hr)) goto again;
	}


	if (SUCCEEDED(hr)) {
		MouDeltaX = dims.lX;
		MouDeltaY = dims.lY;

		MouRelease=0;
		if (MouPress) MouRelease=MouPress;
		MouPress=(dims.rgbButtons[0]&0x80) ? 1:0;
		MouPress|=(dims.rgbButtons[1]&0x80) ? 2:0;
		if (MouPress) MouRelease=0;
	}
}


void DI_GetKeyboardData(void)
{
	HRESULT hr;

	if (!pKeyboard) return;

again:;
	hr = pKeyboard->GetDeviceState(sizeof(Keys), Keys);
	if (hr == DIERR_INPUTLOST) {
		hr = pKeyboard->Acquire();
		if (SUCCEEDED(hr)) goto again;
	}


	if (SUCCEEDED(hr)) {
	}
}



extern "C" {

void DI_SyncAcquire(void)
{
        if (!pMouse||!pKeyboard) return;
        if (Paused) {
                pMouse->Unacquire();
                pKeyboard->Unacquire();
        }
        else {
                pMouse->Acquire();
                pKeyboard->Acquire();
        }
}

}

