#include "engine.h"
#include "h_timer.h"


#define WM_SYNCACQUIRE	(WM_USER + 0)	// Acquire message


int Done=0, Paused=0;

int TotalFrames=0;
int TotalTicks=0, FrameTicks=0;
static int StartTicks=0;


static ClientFrame LocalClient;
static int client_inited=0, server_inited=0;

int LocalPlayer=-1;


extern void DI_SyncAcquire(void);
void ProcessWinMessages(void);



void Engine_Init(EngineImport *eimp)
{
	char str[256];

	sprintf(str,"POLY ENGINE version %s\nWritten by Alexey Goloshubin",
			POLY_VERSION);
	MessageBox(NULL, str, "Poly Engine demo", MB_OK);

	if (!DAT_Open(eimp->dataPath)) {
		sprintf(str,"Could not load pak: %s",eimp->dataPath);
		Engine_Error(str);
	}

// Init network support
	InitNetwork();

// Init AI progs
	InitProgs();

// Init the server
	if (eimp->mode&ENGINE_SERVER) 
		server_inited=Server_Init();

/*	if (!server_inited) {
		client=TRUE;
		if (!Client_Connect()) {
			printf("Could not connect to server.\n");
			server_inited=Server_Init();
		}
	}
*/

	if (eimp->mode&ENGINE_CLIENT) 
		client_inited=Client_Init(eimp);

	if (!client_inited) Engine_FatalError("Client not inited");

	Timer_Init(TIMER_SPEED);

// Create a local player if server and client are on the same machine
	if (client_inited&&server_inited)
		LocalPlayer=World_AddPlayer("Alexey");

}

int  Engine_IsServer(void) {return server_inited;}
int  Engine_IsClient(void) {return client_inited;}



void Engine_Shut()
{
	if (client_inited) Client_Shut();
	if (server_inited) Server_Shut();

	Timer_Shut();
	ShutNetwork();
	DAT_Close();
}

void Engine_NextFrame()
{
	int i;

	if (StartTicks==0) {
		StartTicks=Timer_Read();
		FrameTicks=0; 
		TotalTicks=0; 
		TotalFrames=0;
	}
	else {
		i=Timer_Read()-StartTicks;
		FrameTicks=i-TotalTicks;
		TotalTicks=i;
		TotalFrames++;
	}

	if (client_inited) {
		Client_GetInput(&LocalClient);
		if (!server_inited)
			Client_SendUpdateToServer(&LocalClient);
		else
			Server_ProcessLocalInput(&LocalClient);
	}
	else {
		if (kbhit()) {if (getch()==27) Done=1;}
	}

	if (server_inited) {
		Server_Listen();
		Server_UpdateWorld();
	// Note that LocalClient will not be filled out
	// if there is no LocalPlayer
		Server_MakeFrames(&LocalClient);
	}

	if (client_inited) {
		if (!server_inited)
			Client_Listen(&LocalClient);
		Client_ProcessFrame(&LocalClient);
	}
	else {
//		delay(30);
	}

// Process Windows messages
	ProcessWinMessages();
}

void ProcessWinMessages(void)
{
	MSG msg;

	if (Paused) WaitMessage();

	while(PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) {

	// A false message is a quit
		if (!(GetMessage(&msg, NULL, 0, 0))) return;

	// SyncAcquire
		if (msg.message == WM_SYNCACQUIRE)
			DI_SyncAcquire();

		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
}


/**
 Error messages
*/
void Engine_Error(char *str)
{
		MessageBox(NULL, str, "Poly ERROR", MB_OK);
}

void Engine_FatalError(char *str)
{
		//Engine_Shut();
		MessageBox(NULL, str, "Fatal Poly ERROR", MB_OK);
		Done=1;
}


/**
 Timer system
*/

void Timer_Init(int rate) {}

void Timer_Shut(void) {}

int Timer_Read(void)
{
        return(GetTickCount()/10);
}
