
//#define WIN32_LEAN_AND_MEAN

//#define _STATIC_MESA

#include <stdio.h>
#include <windows.h>
#include <math.h>

#include "GL\gl.h" //"c:\alexey\mesa\mesa-2.6\include\GL\gl.h"
#include "GL\glu.h" //"c:\alexey\mesa\mesa-2.6\include\GL\glu.h"

//#include <d:\oglsdk\include\GL\gl.h>
//#include <d:\oglsdk\include\GL\glu.h>
#include <ddraw.h>
#include <dinput.h>


#define WM_SYNCACQUIRE      (WM_USER + 0)       // Acquire message


extern "C" {

        #include "..\engine.h"
        #include "..\h_input.h"
        #include "..\h_gfx.h"

        extern char start_pak[];
        extern int start_width, start_height;

        extern int do_sound, do_kbd, do_timer;
        extern int do_flip, do_vbe, do_half, do_tex;
}



int ProcessMessages(void);


// DirectDraw stuff

BOOL    DD_Init(HINSTANCE hInstance, int iCmdShow);
void    DD_End(void);
HRESULT DD_PalInit(void);
void    DD_SetPalette(byte *pPal);
BOOL    DD_LockSurface(void);
void    DD_UnlockSurface(void);
void    DD_FlipSurface(void);


// DirectInput stuff (must be called after DD_Init)
BOOL    DI_Init(void);
void    DI_End(void);
void    DI_GetMouseData(void);
void    DI_GetKeyboardData(void);
void    DI_SyncAcquire(void);





// General win stuff
char szAppName[] = "WinPoly";
HINSTANCE               g_hInstance;
HWND                    g_hwnd;

HINSTANCE main_hInstance, main_hPrevInstance;
int main_iCmdShow;

// OpenGL stuff
HDC                     g_hdc;
HGLRC                   g_hrc;

// DirectDraw stuff
DDSURFACEDESC           g_ddsd;
int                     g_ScreenXSize;
int                     g_ScreenYSize;
int                     g_Fullscreen=TRUE;

PALETTEENTRY            g_Palette[256];
LPDIRECTDRAW            g_DD            = NULL;
LPDIRECTDRAWSURFACE     g_Primary       = NULL;
LPDIRECTDRAWSURFACE     g_Back          = NULL;
LPDIRECTDRAWPALETTE     g_DDPalette = NULL;

byte*                   g_screen;
int                     g_row_width;

// DirectInput stuff

LPDIRECTINPUT           g_pdi = NULL;
LPDIRECTINPUTDEVICE     g_pMouse = NULL;
LPDIRECTINPUTDEVICE     g_pKeyboard = NULL;
BOOL                    g_fPaused = FALSE;







char *cmd_ptr;

void cmd_init(char *params)
{
        cmd_ptr=params;
}

char *cmd_next() {
        char *ptr=cmd_ptr;

        if (cmd_ptr!=0) {
                while(cmd_ptr[0]!='\0'&&cmd_ptr[0]!=' ')
                        cmd_ptr++;
                if (cmd_ptr[0]=='\0') {
                        if (ptr==cmd_ptr) ptr=0;
                        cmd_ptr=0;
                }
                else {cmd_ptr[0]='\0';cmd_ptr++;}
        }

        return(ptr);
}



int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int iCmdShow)
{
        char *param;
        char bsp_name[64]="base1.bsp";
		char str[64];

		sprintf(str,"POLY ENGINE version %s\nWritten by Alexey Goloshubin",POLY_VERSION);
		MessageBox(NULL, str, "Poly Engine demo", MB_OK);

        printf("\nPOLY ENGINE version %s\n",POLY_VERSION);
        printf("Written by Alexey Goloshubin\n\n");

        start_width=640; start_height=480;

        cmd_init(szCmdLine);
        while((param=cmd_next())!=0) {
                if (strcmp(param,"-pak")==0) {
                        if ((param=cmd_next())!=0)
                                strcpy(start_pak,param);
                }
                else if (strcmp(param,"-win")==0) {
                        do_vbe=1;
                        if ((param=cmd_next())!=0)
                                sscanf(param,"%d",&start_width);
                        if ((param=cmd_next())!=0)
                                sscanf(param,"%d",&start_height);
                }
                else if (strcmp(param,"-gfx")==0) {
                        if ((param=cmd_next())!=0)
                                sscanf(param,"%d",&start_width);
                        if ((param=cmd_next())!=0)
                                sscanf(param,"%d",&start_height);
                }
                else if (strcmp(param,"-nosound")==0) {
                        do_sound=0;
                }
                else if (strcmp(param,"-flat")==0) {
                        do_tex=0;
                }
                else if (param[0]!='-') {
                        strcpy(bsp_name,param);
                }
        }



// Store win stuff for later init
        main_hInstance = hInstance;
        main_hPrevInstance = hPrevInstance;
        main_iCmdShow = iCmdShow;


// Init the engine
        Engine_Init(1,1);

// !!! If you've setup a server you must call this !!!
        if (Engine_IsServer()) Server_NewWorld(bsp_name);

        while(!Done) {
                Engine_NextFrame();
                ProcessMessages();
        }

        Engine_Shut();

        if (TotalFrames>10) {
                printf("\nTotal frames:       %d\n",TotalFrames);
                printf("Frame rate:           %4.3f\n",
                        (float)(TotalFrames*100)/(float)TotalTicks);
                printf("Time per frame (ms):  %4.3f\n",
                        (float)(TotalTicks*10)/(float)TotalFrames);
        }

        return(0);
}


// xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
//
// ProcessMessages - Process and dispatch any messages in the queue
//

int ProcessMessages(void)
{
        MSG msg;

        if (g_fPaused) WaitMessage();

        while(PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) {
        // A false message is a quit
                if (!(GetMessage(&msg, NULL, 0, 0)))
                        return msg.wParam;
		// SyncAcquire
                if (msg.message == WM_SYNCACQUIRE)
                        DI_SyncAcquire();

				TranslateMessage(&msg);
                DispatchMessage(&msg);
        }
        return(-1);
}



/****************************************************************************
                         OPENGL  STUFF
*****************************************************************************/

void GLResize(int w,int h)
{
        glViewport(0, 0, w, h);
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        gluPerspective(60.0, (GLfloat)w/(GLfloat)h, 2.0, 4000.0);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
}


LRESULT CALLBACK GLWndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
        PAINTSTRUCT ps;
//        RECT rect;
        LRESULT lRc;

        switch (message) {
                case WM_CREATE:
/*                      g_hdc = GetDC(hwnd);
                        if (!bSetupPixelFormat(g_hdc))
                                PostQuitMessage (0);

                        g_hrc = wglCreateContext(g_hdc);
                        wglMakeCurrent(g_hdc, g_hrc);
*/
                        break;

                case WM_PAINT:
                        BeginPaint(hwnd, &ps);
                        EndPaint(hwnd, &ps);
                        break;

                case WM_SIZE:
//3dfx                        GetClientRect(hwnd, &rect);
//                        GLResize(rect.right-rect.left,rect.bottom-rect.top);
                        break;

                case WM_CLOSE:
                        Done=1;
                        if (g_hrc) wglDeleteContext(g_hrc);
                        if (g_hdc) ReleaseDC(hwnd, g_hdc);
                        g_hrc = 0;
                        g_hdc = 0;

                        DestroyWindow (hwnd);
                        break;

                case WM_DESTROY:
                        Done=1;
                        if (g_hrc) wglDeleteContext(g_hrc);
                        if (g_hdc) ReleaseDC(hwnd, g_hdc);
                        g_hrc = 0;
                        g_hdc = 0;

                        PostQuitMessage (0);
                        break;

                case WM_KEYDOWN:
                        switch(wParam) {
                        // Shift ESCAPE throws us out, bugger leaking memory!
                                case VK_ESCAPE:
//                                      if ( GetKeyState( VK_SHIFT) < 0 ) {
                                                PostMessage(hwnd, WM_CLOSE, 0, 0);
//                                      }
                                        break;
                        }
                        break;

                case WM_ACTIVATE:
                        g_fPaused = (wParam==WA_INACTIVE);
                        PostMessage(hwnd, WM_SYNCACQUIRE, 0, 0L);
//                      DI_SyncAcquire();
                        break;

                case WM_ENTERMENULOOP:
                case WM_ENTERSIZEMOVE:
                        g_fPaused = TRUE;
                        DI_SyncAcquire();
                        break;

                case WM_EXITMENULOOP:
                case WM_EXITSIZEMOVE:
                        g_fPaused = GetActiveWindow() != hwnd || IsIconic(hwnd);
                        PostMessage(hwnd, WM_SYNCACQUIRE, 0, 0L);
                        break;

                case WM_SYNCACQUIRE:
                        //DI_SyncAcquire();
                        break;

                case WM_SYSCOMMAND:
                // Ignore ALT - menu key
                        if (wParam==SC_KEYMENU) return 0;

                // Some weird stuff for mouse
                        lRc = DefWindowProc(hwnd, message, wParam, lParam);
                        if (IsWindow(hwnd))
                                DI_SyncAcquire();
                        return lRc;

        }    

        return DefWindowProc(hwnd, message, wParam, lParam);
}


BOOL bSetupPixelFormat(HDC hdc)
{
        PIXELFORMATDESCRIPTOR pfd, *ppfd;
        int pixelformat;

        ppfd = &pfd;

        ppfd->nSize = sizeof(PIXELFORMATDESCRIPTOR);
        ppfd->nVersion = 1;
        ppfd->dwFlags = PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
        ppfd->dwLayerMask = PFD_MAIN_PLANE;
        ppfd->iPixelType = PFD_TYPE_RGBA;
        ppfd->cColorBits = 16;
        ppfd->cDepthBits = 16;
        ppfd->cAccumBits = 0;
        ppfd->cStencilBits = 0;

        pixelformat = ChoosePixelFormat(hdc, ppfd);

        if ((pixelformat = ChoosePixelFormat(hdc, ppfd)) == 0) {
                MessageBox(NULL, "ChoosePixelFormat failed", "Error", MB_OK);
                return FALSE;
        }

        if (pfd.dwFlags & PFD_NEED_PALETTE) {
                MessageBox(NULL, "Needs palette", "Error", MB_OK);
                return FALSE;
        }

        if (SetPixelFormat(hdc, pixelformat, ppfd) == FALSE) {
                MessageBox(NULL, "SetPixelFormat failed", "Error", MB_OK);
                return FALSE;
        }

        return TRUE;
}


BOOL GL_Init(int width, int height, int flags)
{
        WNDCLASS wc;
        HWND hwnd;
        HRESULT ddrval;
        RECT rect;

        wc.style         = 0;
        wc.lpfnWndProc   = (WNDPROC)GLWndProc;
        wc.cbClsExtra    = 0;
        wc.cbWndExtra    = 0;
        wc.hInstance     = main_hInstance;
        wc.hIcon         = LoadIcon (main_hInstance, szAppName);
        wc.hCursor       = LoadCursor (NULL,IDC_ARROW);
        wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
        wc.lpszMenuName  = szAppName;
        wc.lpszClassName = szAppName;

        if (!RegisterClass(&wc)) return FALSE;

        if (flags&GFX_WINDOWED) g_Fullscreen=FALSE;
        else g_Fullscreen=TRUE;

        if (g_Fullscreen) {
        // Dummy fullscreen window
                hwnd = CreateWindow(szAppName, "Poly/OpenGL", 
                        WS_POPUP | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
                        0, 0,
                        width, height,
                        NULL, NULL, main_hInstance, NULL);
        }
        else {
        // Normal window
                hwnd = CreateWindow(szAppName, "Poly/OpenGL", 
                        WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_VISIBLE| WS_OVERLAPPEDWINDOW,
                        CW_USEDEFAULT,CW_USEDEFAULT,
                        width, height,
                        NULL, NULL, main_hInstance, NULL);
        }

        if (!hwnd) return FALSE;

        g_hwnd  = hwnd;
        g_hInstance = main_hInstance;


        if (g_Fullscreen) {
                SetCursor(NULL);
                ShowCursor(NULL);
        }

        ShowWindow(hwnd, main_iCmdShow);
        UpdateWindow(hwnd);


// Use DD to change display mode for fullscreen
        if (g_Fullscreen) {
                ddrval = DirectDrawCreate(NULL, &g_DD, NULL);
                if (ddrval!=DD_OK) goto error;

                ddrval=g_DD->SetCooperativeLevel(hwnd,(DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN));
                if (ddrval != DD_OK) goto error;

                ddrval = g_DD->SetDisplayMode(width, height, 16);
                if (ddrval != DD_OK ) goto error;

        // Workaround for RIVA 128
                ddrval=g_DD->SetCooperativeLevel(hwnd,(DDSCL_NORMAL|DDSCL_FULLSCREEN));
                if (ddrval != DD_OK) goto error;
        }

// Setup GL context
        g_hdc = GetDC(g_hwnd);
        if (!bSetupPixelFormat(g_hdc)) goto error;

        g_hrc = wglCreateContext(g_hdc);
        wglMakeCurrent(g_hdc, g_hrc);

        GetClientRect(hwnd, &rect);
        GLResize(rect.right-rect.left,rect.bottom-rect.top);

        return TRUE;

error:

        printf("Error in GL init %d\n",ddrval);

        DD_End();
        DestroyWindow(hwnd);
        return FALSE;
}






/****************************************************************************
                         DIRECTX  STUFF
*****************************************************************************/


LRESULT CALLBACK DDWndProc(HWND hwnd, UINT message , WPARAM wParam, LPARAM lParam)
{
        LRESULT lRc;
        
        switch(message) {

                case WM_CREATE:
                        break;

                case WM_ACTIVATE:
                        g_fPaused = (wParam==WA_INACTIVE);
                        DI_SyncAcquire();
                        break;

                case WM_ENTERMENULOOP:
                case WM_ENTERSIZEMOVE:
                        g_fPaused = TRUE;
                        DI_SyncAcquire();
                        break;

                case WM_EXITMENULOOP:
                case WM_EXITSIZEMOVE:
                        g_fPaused = GetActiveWindow() != hwnd || IsIconic(hwnd);
                        PostMessage(hwnd, WM_SYNCACQUIRE, 0, 0L);
                        break;

                case WM_SYNCACQUIRE:
                        //DI_SyncAcquire();
                        break;

                case WM_SYSCOMMAND:
                // Ignore ALT - menu key
                        if (wParam==SC_KEYMENU) return 0;

                // Some weird stuff for mouse
                        lRc = DefWindowProc(hwnd, message, wParam, lParam);
                        if (IsWindow(hwnd))
                                DI_SyncAcquire();
                        return lRc;

                case WM_KEYDOWN:
                        switch(wParam) {
                        // Shift ESCAPE throws us out, bugger leaking memory!
                                case VK_ESCAPE:
//                                      if ( GetKeyState( VK_SHIFT) < 0 ) {
                                                PostMessage(hwnd, WM_CLOSE, 0, 0);
                                                printf("escape\n");
//                                      }
                                        break;
                        }
                        break;

                case WM_SETCURSOR:
//                      if (g_Fullscreen)
                                SetCursor(NULL);
                                ShowCursor(NULL);
                        return TRUE;

                case WM_CLOSE:
printf("close\n");
                        Done=1;
                        DestroyWindow (hwnd);
                        break;

                case WM_DESTROY:
printf("destroy\n");
                        Done=1;
                        DD_End();
                        PostQuitMessage(0);
                        break;
        }

        return DefWindowProc(hwnd, message, wParam, lParam);
}



// xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
//
// DD_Init - Create a window and initialize DirectDraw
//

BOOL DD_Init(int width, int height, int flags)
{
        HRESULT                 ddrval;
        HWND                    hwnd;
        WNDCLASS                wc;
        DDSURFACEDESC           ddsd;
        char                    buffer[256];
        DDSCAPS                 ddscaps;

// Always full screen
        g_Fullscreen=TRUE;

        wc.style         = CS_HREDRAW | CS_VREDRAW;
        wc.lpfnWndProc   = (WNDPROC) DDWndProc;
        wc.cbClsExtra    = 0;
        wc.cbWndExtra    = 0;
        wc.hInstance     = main_hInstance;
        wc.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
        wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
        wc.hbrBackground = NULL;                //      (HBRUSH)GetStockObject(WHITE_BRUSH);
        wc.lpszMenuName  = NULL;                //      MAKEINTRESOURCE(IDR_MAINMENU);
        wc.lpszClassName = szAppName;
        RegisterClass(&wc);


// Dummy fullscreen window
        hwnd = CreateWindow(szAppName, "WinPoly", 
                WS_POPUP | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
                0, 0,
                width, height,
                NULL, NULL, main_hInstance, NULL);

        if (!hwnd) return FALSE;

        g_hwnd  = hwnd;
        g_hInstance = main_hInstance;

        // Get rid of the Bloomin' system mouse pointer!!!

        SetCursor(NULL);
        ShowCursor(NULL);

        ShowWindow( hwnd, main_iCmdShow );
        UpdateWindow( hwnd );

        ddrval = DirectDrawCreate(NULL, &g_DD, NULL);
        if (ddrval!=DD_OK) goto error;

        ddrval=g_DD->SetCooperativeLevel(hwnd,(DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN));
        if (ddrval!=DD_OK) goto error;

        ddrval = g_DD->SetDisplayMode(width, height, 8);
        if (ddrval!=DD_OK) goto error;

        ddrval = DD_PalInit();
        if ( ddrval != DD_OK ) goto error;

        g_ScreenXSize = width;
        g_ScreenYSize = height;

// Create the Primary buffer and back buffer.
        ddsd.dwSize  = sizeof(ddsd);
        ddsd.dwFlags =
                DDSD_CAPS |
                DDSD_BACKBUFFERCOUNT;

        ddsd.ddsCaps.dwCaps =
                DDSCAPS_PRIMARYSURFACE |
                DDSCAPS_FLIP |
                DDSCAPS_COMPLEX;

        ddsd.dwBackBufferCount = 1;

        ddrval=g_DD->CreateSurface(&ddsd,&g_Primary,NULL);
        if (ddrval!=DD_OK) goto error;

// Get back buffer pointer
        ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
        ddrval = g_Primary->GetAttachedSurface(&ddscaps, &g_Back);
        if (ddrval!=DD_OK) goto error;

// Workaround for RIVA 128
//      ddrval=g_DD->SetCooperativeLevel(hwnd,(DDSCL_NORMAL|DDSCL_FULLSCREEN));
//      if (ddrval!=DD_OK) goto error;

        return TRUE;
// Error!

error:

        wsprintf(buffer, "Error : %d", ddrval);

        MessageBox(hwnd, buffer, "* DIRECT X ERROR *", MB_OK);

        DD_End();
        DestroyWindow(hwnd);
        return FALSE;
}


// xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
//
// DD_End - Close down DirectDraw
//

void DD_End(void)
{
        if (g_DD!=NULL) {
                if (g_Primary!=NULL) {
                        g_Primary->Release();
                        g_Primary = NULL;
                }

                if (g_DDPalette!=NULL) {
                        g_DDPalette->Release();
                        g_DDPalette = NULL;
                }
                g_DD->Release();
                g_DD = NULL;
        }
}

// xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
//
// DD_PalInit - Create a DirectDraw 8 bit palette
//

HRESULT DD_PalInit(void)                // Setup the palette
{
        HRESULT ddrval;

        if (g_DD!=NULL) {
                ddrval=g_DD->CreatePalette(
                        DDPCAPS_8BIT | DDPCAPS_ALLOW256,
                        g_Palette, &g_DDPalette, NULL);

                if (g_DDPalette!=NULL) {
                        if (g_Primary!=NULL) {
                                ddrval = g_Primary->SetPalette(g_DDPalette);
                        }
                }
        }

        return(ddrval);
}

// xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
//
// DD_SetPalette - Set the 8 bit DDRAW palette
//

void DD_SetPalette(byte *pPal)
{
        int i;

        for (i = 0; i < 256; i++)
        {
                g_Palette[i].peRed   = pPal[i*3+0];
                g_Palette[i].peGreen = pPal[i*3+1];
                g_Palette[i].peBlue  = pPal[i*3+2];
                g_Palette[i].peFlags = 0;
        }

        if (g_DDPalette)
        {
                g_DDPalette->SetEntries(0, 0, 256, g_Palette);
                g_Primary->SetPalette(g_DDPalette);
        }
}

// xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
//
// LockSurface - Lock the drawing surface, point g_screen to the surface.
//

BOOL DD_LockSurface(void)
{
        BOOL    r;

        r = FALSE;
        if (g_DD != NULL) {
                if ( g_Primary->IsLost() ) {
                        g_Primary->Restore();
                }

                g_ddsd.dwSize = sizeof(g_ddsd);

                if (g_Back->Lock(NULL,&g_ddsd,DDLOCK_SURFACEMEMORYPTR|DDLOCK_WAIT,NULL)==DD_OK) {
                        g_screen = (byte *)g_ddsd.lpSurface;
                        g_row_width = g_ddsd.lPitch;
                        r = TRUE;
                }
        }
        return(r);
}

void DD_UnlockSurface(void)
{
        if (g_Back==NULL) return;
        g_Back->Unlock(NULL);
}

// xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
//
// FlipSurface - Flip the screens
//

void DD_FlipSurface(void)
{
        HRESULT                 ddrval;

        while(1) {

                ddrval = g_Primary->Flip(NULL,0);
                if (ddrval==DD_OK) break;

                if (ddrval==DDERR_SURFACELOST) {
                        ddrval = g_Primary->Restore();
                        if (ddrval!=DD_OK) break;
                }
                if (ddrval!=DDERR_WASSTILLDRAWING) break;
        }
}



/****************************************************************************
                        DIRECT INPUT CODE
*****************************************************************************/


// Init all DI devices (mouse, kbd). Must be called after either DD_Init() or
// GL_Init()

BOOL DI_Init(void)
{
        HRESULT hr;

// Already initialized
        if (g_pdi) return TRUE;

// Init DirectInput
        hr = DirectInputCreate(g_hInstance, DIRECTINPUT_VERSION, &g_pdi, NULL);
        if (FAILED(hr)) return FALSE;

// Init mouse
        hr = g_pdi->CreateDevice(GUID_SysMouse, &g_pMouse, NULL);
        if (FAILED(hr)) return FALSE;

        hr = g_pMouse->SetDataFormat(&c_dfDIMouse);
        if (FAILED(hr)) return FALSE;

        hr = g_pMouse->SetCooperativeLevel(g_hwnd,DISCL_EXCLUSIVE | DISCL_FOREGROUND);
        if (FAILED(hr)) return FALSE;

// Init keyboard
        hr = g_pdi->CreateDevice(GUID_SysKeyboard, &g_pKeyboard, NULL);
        if (FAILED(hr)) return FALSE;

        hr = g_pKeyboard->SetDataFormat(&c_dfDIKeyboard);
        if (FAILED(hr)) return FALSE;

        hr = g_pKeyboard->SetCooperativeLevel(g_hwnd,DISCL_NONEXCLUSIVE | DISCL_FOREGROUND);
        if (FAILED(hr)) return FALSE;


printf("DI init ok\n");

// Since we already have a window, acquire mouse and kbd
        g_fPaused=FALSE;
        DI_SyncAcquire();

        return TRUE;
}


void DI_End(void)
{

// Shut keyboard
        if (g_pKeyboard) {
                g_pKeyboard->Unacquire();
                g_pKeyboard->Release();
                g_pKeyboard = NULL;
        }

// Shut mouse
        if (g_pMouse) {
                g_pMouse->Unacquire();
                g_pMouse->Release();
                g_pMouse = NULL;
        }

// Shut DirectInput
        if (g_pdi) {
                g_pdi->Release();
                g_pdi = NULL;
        }
}


void DI_GetMouseData(void)
{
        DIMOUSESTATE dims;
        HRESULT hr;

        if (!g_pMouse) return;


again:;
        hr = g_pMouse->GetDeviceState(sizeof(DIMOUSESTATE), &dims);
        if (hr == DIERR_INPUTLOST) {
                hr = g_pMouse->Acquire();
                if (SUCCEEDED(hr)) goto again;
        }


        if (SUCCEEDED(hr)) {

                MouDeltaX = dims.lX;
                MouDeltaY = dims.lY;

                MouRelease=0;
                if (MouPress) MouRelease=MouPress;
                MouPress=(dims.rgbButtons[0]&0x80) ? 1:0;
                MouPress|=(dims.rgbButtons[1]&0x80) ? 2:0;
                if (MouPress) MouRelease=0;
        }
}


void DI_GetKeyboardData(void)
{
        HRESULT hr;

        if (!g_pKeyboard) return;


again:;
        hr = g_pKeyboard->GetDeviceState(sizeof(Keys), Keys);
        if (hr == DIERR_INPUTLOST) {
                hr = g_pKeyboard->Acquire();
                if (SUCCEEDED(hr)) goto again;
        }


        if (SUCCEEDED(hr)) {
        }
}


void DI_SyncAcquire(void)
{
        if (!g_pMouse||!g_pKeyboard) return;
        if (g_fPaused) {
                g_pMouse->Unacquire();
                g_pKeyboard->Unacquire();
        }
        else {
                g_pMouse->Acquire();
                g_pKeyboard->Acquire();
        }
}







/****************************************************************************
                GFX, INPUT and TIMER systems for POLY Engine
*****************************************************************************/


extern "C" {

// Gfx video info
int gfxScrWidth, gfxScrHeight;  // Screen width and height
int gfxScanWidth;               // Scanline width

// Mouse input
int MouX, MouY, MouDeltaX, MouDeltaY, MouPress, MouRelease;

// Kbd input
byte Keys[256], KeyPressed;

// VR stuff (placeholder)
int VR_Yaw=0, VR_Pitch=0, VR_Roll=0;


void GFX_InitGraph(int w,int h,int flags)
{
        int a;

        if (!DD_Init(w,h,flags))        // Init DDraw
                Engine_FatalError("Could not init DirectDraw");

        if (DD_LockSurface()) {
                // Only unlock surface if locked OK
                g_Back->Unlock(g_ddsd.lpSurface);
        }
        a = ProcessMessages();
//      if (a!=-1)
//              return(a);

        gfxScrWidth=g_ScreenXSize;
        gfxScrHeight=g_ScreenYSize;
        gfxScanWidth=g_row_width;

        printf("DDraw: %d %d %d\n",g_ScreenXSize,g_ScreenYSize,g_row_width);
}


byte *GFX_LockSurface(int page)
{
        DD_LockSurface();
        return((byte *)g_screen);
}

void GFX_UnlockSurface(void)
{
        DD_UnlockSurface();
}

void GFX_FlipSurface(void)
{
        DD_FlipSurface();
}


void GFX_ShutGraph(void)
{
        DD_End();
        PostMessage(g_hwnd, WM_CLOSE, 0, 0);
}

void GFX_SetPalette(byte *rgbpal)
{
        DD_SetPalette(rgbpal);
}


void Engine_FatalError(char *str)
{
        Engine_Shut();
        if (str) printf("\nERROR: %s\n\n",str);
        else printf("\nFATAL ERROR\n\n");

        MessageBox(g_hwnd, str, "Fatal Poly ERROR", MB_OK);
        DestroyWindow(g_hwnd);
}



void Timer_Init(int rate) {}

void Timer_Shut(void) {}


int Timer_Read(void)
{
        return(GetTickCount()/10);
}


int Input_Init(int flags)
{
        return (DI_Init());
}

void Input_Shut(void)
{
        DI_End();
}

int Input_GetData(void)
{
        DI_GetMouseData();
        DI_GetKeyboardData();
        return(INPUT_KEYBOARD|INPUT_MOUSE);
}


void OGL_Init(int width, int height, int flags) {
        if (!GL_Init(width, height, flags))
                Engine_FatalError("Could not init OpenGL");
}

void OGL_Shut() {
        DD_End();
}

void OGL_Flip() {
        SwapBuffers(g_hdc);
}

int OGL_GetWidth(void) {
        RECT rect;
        GetClientRect(g_hwnd, &rect);
        return(rect.right-rect.left);
}

int OGL_GetHeight(void) {
        RECT rect;
        GetClientRect(g_hwnd, &rect);
        return(rect.bottom-rect.top);
}

}

