/********************************************************************
 NETGAME.C - Network code (high level) for the engine
*********************************************************************/

#include "engine.h"
#include "progs\progs.h"
#include "h_net.h"

//#include <i86.h>		// for delay()

#define NET_VERSION	0x0103	// 1.3


/*
 Compressed visible object (for client update)
*/
typedef struct {
	short c[3];
	byte angle;
	byte vangle;
	byte prog_idx;
	byte frame;
	byte flags;
} XPOB;


/*
 Connection information
*/
typedef struct {
	int  device;
	byte address[6];
	int  connect;
	int  timeout;
} Connect;

Connect Connects[MAX_PLAYERS];

char ComputerName[16]="TheServer";



/************************************************************************
			PACKETS STUFF
*************************************************************************/

enum {
	PK_ACK=1,
	PK_END,
	PK_REQ_SERVER_INFO,
	PK_SERVER_INFO,
	PK_REQ_CONNECT,
	PK_PLAYER_INFO,
	PK_GAME_INFO,
	PK_UPDATE_SERVER,
	PK_UPDATE_CLIENT,
};


typedef struct {
	short Version;
	char  Name[16];
	char  MapName[16];
	byte  NumPlayers;
	byte  MaxPlayers;
} PK_ServerInfo;

typedef struct {
	char Name[16]; 
} PK_PlayerInfo;

typedef struct {
	char MapName[64];
} PK_GameInfo;

typedef struct {
	float cam_pos[3];
	float cam_angle;
	float body_roll;
	float cam_height;
	int flags;
	int NumVisObjects;	// Followed by XPOB*NumVisObjects
} PK_UpdateClient;

typedef struct {
	int action;
	float body_yaw;
	float head_yaw;
	float head_pitch;
} PK_UpdateServer;


typedef struct {
	int type; 
} PK_Header;


static byte NetBuffer[512];
static PK_Header *NetHeader=(PK_Header *)NetBuffer;
static byte *NetData=NetBuffer+sizeof(PK_Header);

static int MaxXPOBs=(512-sizeof(PK_Header)-sizeof(PK_UpdateClient))
		/sizeof(XPOB);



int Client_SelectServer(int *pdevice,byte *paddress);
void Client_ProcessNewWorld(void);
void Client_ProcessUpdate(ClientFrame *cl);
int GetPlayerFromAddress(int device, byte *adr);
void Server_SendInfo(PktInfo *pkt);
void Server_ProcessUpdate(int player_idx);







int InitNetwork(void)
{
	FILE *fp;
	char buf[80], d[8];
	int status=1,temp,p1,p2;

	fp=fopen("NET.INI","rt");
	if (!fp) {
		printf("Could not open NET.INI\n");
		return 0;
	}

	while(fgets(buf,78,fp)) {
		if (buf[0]==';'||buf[0]=='\r'||buf[0]=='\n') continue;
		sscanf(buf,"%s",d);

		if (stricmp(d,"NAME")==0) {
			sscanf(buf,"%s %s",d,ComputerName);
		}
		else if (strnicmp(d,"COM",3)==0) {
			p1=(d[3]-'0');
			if (p1<NDEV_COM1||p1>NDEV_COM4) continue;
			sscanf(buf,"%s %d",d,&p2);
			temp=Net_InitDevice(p1,0x6576,p2);
			strupr(d);
			printf("NetDevice: %s (%d) - %s\n",
				d,p2,temp?"ON":"OFF");
			status|=temp;
		}
		else if (stricmp(d,"IPX")==0) {
			temp=Net_InitDevice(NDEV_IPX,0x6576,0);
			strupr(d);
			printf("NetDevice: %s - %s\n",d,temp?"ON":"OFF");
			status|=temp;
		}
	}
	fclose(fp);

	printf("Net protocol version %d.%d\n",NET_VERSION>>8,NET_VERSION&0xFF);

	return(status);
}

void ShutNetwork(void)
{
	Net_ShutAllDevices();
}

int Client_Connect(void)
{
	PK_PlayerInfo *pinfo;
	PK_GameInfo *ginfo;
	PktInfo server, pkt;
	int done;

	if (!Client_SelectServer(&server.device,server.address)) return 0;

// Send REQ_CONNECT
	NetHeader->type=PK_REQ_CONNECT;
	server.size=sizeof(PK_Header);
	Net_SendPacket(&server,NetBuffer);
	printf("send REQ_CONNECT\n");

// Wait for ACK from the selected server
	pkt=server;
	done=0;
	while(!kbhit()&&!done) {
		pkt.size=500;
		if (Net_GetPacket(&pkt,NetBuffer)) {
			if (!memcmp(server.address,pkt.address,6)) {
				if (NetHeader->type==PK_ACK) done=1;
				else if (NetHeader->type==PK_END) break;
			}
		}
	}
// Interrupted or refused?
	if (!done) return 0;
	printf("got  ACK\n");

// Send PLAYER_INFO
	NetHeader->type=PK_PLAYER_INFO;
	pinfo=(PK_PlayerInfo *)NetData;
	strcpy(pinfo->Name,"ThePlayer");
	server.size=sizeof(PK_Header)+sizeof(PK_PlayerInfo);
	Net_SendPacket(&server,NetBuffer);
	printf("send PLAYER_INFO\n");

// Wait for GAME_INFO from our server
	pkt=server;
	done=0;
	while(!kbhit()&&!done) {
		pkt.size=500;
		if (Net_GetPacket(&pkt,NetBuffer)) {
			if (!memcmp(server.address,pkt.address,6)&&
			    NetHeader->type==PK_GAME_INFO) done=1;
		}
	}
// Interrupted?
	if (!done) return 0;
	printf("got  GAME_INFO\n");

// Init game
	ginfo=(PK_GameInfo *)NetData;
	if (!InitWorld(ginfo->MapName)) {
		NetHeader->type=PK_END;
		server.size=sizeof(PK_Header);
		Net_SendPacket(&server,NetBuffer);
		Done=1;
		return 0;
	}

// Set the connection with the server
	Connects[0].device=server.device;
	memcpy(Connects[0].address,server.address,6);
	Connects[0].connect=10;

// Send ACK to the server
	NetHeader->type=PK_ACK;
	server.size=sizeof(PK_Header);
	Net_SendPacket(&server,NetBuffer);
	printf("send ACK\n");

// Success
	printf("Connection established.\n");

	return 1;
}

void Client_Listen(ClientFrame *cl)
{
	PktInfo server,pkt;
	int type;


	server.device=Connects[0].device;
	memcpy(server.address,Connects[0].address,6);

	pkt=server; pkt.size=500;
	while(Net_GetPacket(&pkt,NetBuffer)) {
		if (memcmp(pkt.address,server.address,6)) continue;

		type=NetHeader->type;
		if (type==PK_END) {
			Done=1;
		}
		else if (type==PK_GAME_INFO) {
			Client_ProcessNewWorld();
			cl->req=1;
		}
		else if (type==PK_UPDATE_CLIENT) {
			Client_ProcessUpdate(cl);
			cl->req=1;
		}
		pkt.size=500;
	}
}


void Server_Listen(void)
{
	Player *p;
	PK_PlayerInfo *pinfo;
	PktInfo pkt;
	int type,i,idx,con;
//	char params[32];

// Check for timed out players
	for(i=0;i<NumPlayers;i++) {
		if (i==LocalPlayer) continue;
		p=&Players[i];
		Connects[i].timeout+=FrameTicks;
		if (Connects[i].timeout>500) {
			if (p->obj_idx>=0)
				Prog_Event(p->obj_idx,PE_DEATH,0,0);
			p->obj_idx=-1;
			World_RemovePlayer(i);
			printf("Player %d - timed out\n",i);
		}
	}

// Get packets
	for(i=0;i<NUM_NDEVS;i++) {
		pkt.device=i;
		pkt.size=500;
		while(Net_GetPacket(&pkt,NetBuffer)) {
			type=NetHeader->type;
			idx=GetPlayerFromAddress(pkt.device,pkt.address);
			if (idx<0) con=0;
			else {
				p=&Players[idx];
				con=Connects[idx].connect;
			}

			if (type==PK_REQ_SERVER_INFO) {
				Server_SendInfo(&pkt);
			}
			else if (type==PK_REQ_CONNECT&&con==0) {
				printf("got  REQ_CONNECT\n");
			// If too many players, refuse
				if (NumPlayers>=MAX_PLAYERS) {
					NetHeader->type=PK_END;
					pkt.size=sizeof(PK_Header);
					Net_SendPacket(&pkt,NetBuffer);
					printf("send END\n");
				}
				else {
				// Add new player and start connection
					idx=World_AddPlayer("unknown");
					Connects[idx].device=pkt.device;
					memcpy(Connects[idx].address,
						pkt.address,6);
					Connects[idx].connect=5;
					Connects[idx].timeout=0;
				// Send ACK
					NetHeader->type=PK_ACK;
					pkt.size=sizeof(PK_Header);
					Net_SendPacket(&pkt,NetBuffer);
					printf("send ACK\n");
				}
			}
			else if (type==PK_PLAYER_INFO) {
				printf("got  PLAYER_INFO\n");
				pinfo=(PK_PlayerInfo *)NetData;
				strcpy(p->name,pinfo->Name);

				if (con==5) {
				// Send GAME_INFO
					Server_SendNewWorld(idx);
					Connects[idx].connect=7;
				}
			}
			else if (type==PK_ACK&&con==7) {
				printf("got  ACK\n");
				p->obj_idx=World_AddObject(PROG_PLAYER,NULL);
//sprintf(params,"\"follow\" \"%d\"",Players[0].obj_idx);
//p->camobj_idx=World_AddObject(PROG_EXTCAMERA,params);
				Entity_PlayerStartCoords(&WorldBSP,Objects[p->obj_idx].pob.c);
				World_ObjectMoved(p->obj_idx);
				Connects[idx].connect=10;
				p->req=1;
			}
			else if (type==PK_UPDATE_SERVER) {
				Server_ProcessUpdate(idx);
				Connects[idx].timeout=0;
				p->req=1;
			}
			else if (type==PK_END&&con>0) {
				if (con==10)
					Prog_Event(p->obj_idx,PE_DEATH,0,0);
				p->obj_idx=-1;
				World_RemovePlayer(idx);
			}

			pkt.size=500;
		}
	}
}


/*
 ClientFrame must be properly filled out.
*/
void Server_SendUpdateToClient(int player_idx,ClientFrame *cl)
{
	PktInfo pkt;
	Player *player=&Players[player_idx];
	Connect *con=&Connects[player_idx];
	PK_UpdateClient *upd=(PK_UpdateClient *)NetData;
	XPOB *xpob=(XPOB *)(NetData+sizeof(PK_UpdateClient));
	POB  *pob=cl->VisObjects;
	int num,i;

	if (con->connect!=10) return;
//	if (!player->req) return;

	num=min(MaxXPOBs,cl->NumVisObjects);

// Compress visible objects
	for(i=0;i<num;i++,pob++,xpob++) {
		xpob->c[0]=(short)pob->c[0];
		xpob->c[1]=(short)pob->c[1];
		xpob->c[2]=(short)pob->c[2];
		xpob->angle=(byte)(pob->angle*128.0/PI);
		xpob->vangle=(byte)(pob->vangle*128.0/PI);
		xpob->frame=(byte)pob->frame;
		xpob->prog_idx=(byte)pob->prog_idx;
		xpob->flags=(byte)pob->flags;
	}

// Make the packet
	pkt.device=con->device;
	memcpy(pkt.address,con->address,6);
	pkt.size=sizeof(PK_Header)+sizeof(PK_UpdateClient)+num*sizeof(XPOB);

	NetHeader->type=PK_UPDATE_CLIENT;
	FVec_Copy(cl->cam_pos,upd->cam_pos);
	upd->cam_angle=cl->cam_angle;
	upd->body_roll=player->body_roll;
	upd->cam_height=player->cam_height;
	upd->flags=cl->player_flags;
	upd->NumVisObjects=cl->NumVisObjects;

	Net_SendPacket(&pkt,NetBuffer);

	player->req=0;
}


/*
 NetBuffer contains a packet with visible objects in compressed form
 and the camera position for local player
*/
void Client_ProcessUpdate(ClientFrame *cl)
{
	PK_UpdateClient *upd=(PK_UpdateClient *)NetData;
	XPOB *xpob=(XPOB *)(NetData+sizeof(PK_UpdateClient));
	POB  *pob=cl->VisObjects;
	int i;


	cl->MainBSP=&WorldBSP;

// Read cam position and number of visible objects
	cl->NumVisObjects=upd->NumVisObjects;
	FVec_Copy(upd->cam_pos,cl->cam_pos);
	cl->cam_angle=upd->cam_angle;
	cl->body_roll=upd->body_roll;
	cl->cam_height=upd->cam_height;
	cl->player_flags=upd->flags;

// Decompress XPOBs
	for(i=0;i<cl->NumVisObjects;i++,pob++,xpob++) {
		pob->c[0]=(float)xpob->c[0];
		pob->c[1]=(float)xpob->c[1];
		pob->c[2]=(float)xpob->c[2];
		pob->angle=((float)xpob->angle)*PI/128.0;
		pob->vangle=((float)xpob->vangle)*PI/128.0;
		pob->frame=(int)xpob->frame;
		pob->prog_idx=(int)xpob->prog_idx;
		pob->flags=(int)xpob->flags;
	}
}

/*
  Called only by client. req field indicates that the server
  is ready to accept new coords for the player.
*/
void Client_SendUpdateToServer(ClientFrame *cl)
{
	PktInfo pkt;
	PK_UpdateServer *upd=(PK_UpdateServer *)NetData;
	Connect *con=&Connects[0];

//	if (!cl->req) return;

	pkt.device=con->device;
	memcpy(pkt.address,con->address,6);
	pkt.size=sizeof(PK_Header)+sizeof(PK_UpdateServer);

	NetHeader->type=PK_UPDATE_SERVER;

	upd->action=cl->player_action;
	upd->body_yaw=cl->body_yaw;
	upd->head_yaw=cl->head_yaw;
	upd->head_pitch=cl->head_pitch;


	Net_SendPacket(&pkt,NetBuffer);

	cl->player_action=0;
	cl->req=0;
}


void Server_ProcessUpdate(int player_idx)
{
	PK_UpdateServer *upd=(PK_UpdateServer *)NetData;
	Player *player=&Players[player_idx];
	Object *obj=&Objects[player->obj_idx];

	player->action=upd->action;
	obj->pob.angle+=upd->body_yaw-player->body_yaw;
	player->body_yaw=upd->body_yaw;
	player->head_yaw=upd->head_yaw;
	player->head_pitch=upd->head_pitch;
}


/*
 Send information about this server to some machine.
*/
void Server_SendInfo(PktInfo *pkt)
{
	PK_ServerInfo *info;

	info=(PK_ServerInfo *)NetData;

	NetHeader->type=PK_SERVER_INFO;
	info->Version=NET_VERSION;
	strcpy(info->Name,ComputerName);
	strcpy(info->MapName,WorldBSPName);
	info->NumPlayers=NumPlayers;
	info->MaxPlayers=MAX_PLAYERS;

	pkt->size=sizeof(PK_Header)+sizeof(PK_ServerInfo);
	Net_SendPacket(pkt,NetBuffer);
}


void Server_SendNewWorld(int player_idx)
{
	PktInfo pkt;
	Connect *con=&Connects[player_idx];
	PK_GameInfo *ginfo=(PK_GameInfo *)NetData;

	pkt.device=con->device;
	memcpy(pkt.address,con->address,6);
	NetHeader->type=PK_GAME_INFO;
	pkt.size=sizeof(PK_Header)+sizeof(PK_GameInfo);

	strcpy(ginfo->MapName,WorldBSPName);
	Net_SendPacket(&pkt,NetBuffer);
}

void Client_ProcessNewWorld(void)
{
	PK_GameInfo *ginfo=(PK_GameInfo *)NetData;

	if (!InitWorld(ginfo->MapName))
		Done=1;
}



void Server_Disconnect(void)
{
	PktInfo pkt;
	int i;

	NetHeader->type=PK_END;
	pkt.size=sizeof(PK_Header);

	for(i=0;i<NumPlayers;i++) {
		pkt.device=Connects[i].device;
		memcpy(pkt.address,Connects[i].address,6);
		Net_SendPacket(&pkt,NetBuffer);
	}
//	delay(50);
}

void Client_Disconnect(void)
{
	PktInfo pkt;

	NetHeader->type=PK_END;
	pkt.size=sizeof(PK_Header);

	pkt.device=Connects[0].device;
	memcpy(pkt.address,Connects[0].address,6);
	Net_SendPacket(&pkt,NetBuffer);
//	delay(50);
}


int GetPlayerFromAddress(int device, byte *adr)
{
	Connect *p=Connects;
	int i;

	for(i=0;i<NumPlayers;i++,p++) {
		if (p->device==device&&
		    !memcmp(p->address,adr,6))
			return i;
	}
	return -1;
}



int Client_SelectServer(int *pdevice,byte *paddress)
{
	PktInfo send_pkt, pkt;
	PK_ServerInfo sv_info[16];
	byte sv_address[16][12];
	int sv_device[16];
	int sv_num;
	int i,j,idx,ch;

	printf("\nLooking for servers. Press a key to stop.\n");

	sv_num=0;


	while(!kbhit()) {

	// Broadcast REQ_SERVER_INFO to all devices
		for(i=0;i<NUM_NDEVS;i++) {

	memset(send_pkt.address,0xFF,6);
	send_pkt.size=sizeof(PK_Header);

			send_pkt.device=i;
			NetHeader->type=PK_REQ_SERVER_INFO;
			if (Net_SendPacket(&send_pkt,NetBuffer));// delay(3);
		}
		putch('.');

	// Delay a bit
//		for(i=0;i<100&&!kbhit();i++)
//			delay(10);

	// Get new server info
		for(i=0;i<NUM_NDEVS&&sv_num<16;i++) {
			pkt.device=i;
		// Did we get a packet
			pkt.size=500;
			while(Net_GetPacket(&pkt,NetBuffer)) {
/*
if (i==2)
printf("  %d\n",i);
*/
			// Is this the right type of packet
				if (NetHeader->type!=PK_SERVER_INFO) continue;
			// Check if this server is already on the list
				for(j=0;j<sv_num;j++) {
					if (sv_device[j]==i&&
					  !memcmp(sv_address[j],pkt.address,6))
						break;
				}
				if (j==sv_num) {
				// If we are here, this is a new server
					j=sv_num; sv_num++;
					sv_device[j]=i;
					memcpy(sv_address[j],pkt.address,6);
					memcpy(&sv_info[j],NetData,sizeof(PK_ServerInfo));
					putch('o');
					if (sv_num>=16) {
						ungetch(32);
						break;
					}
				}
				pkt.size=500;
			}
		}
	}
	while(kbhit()) getch();

	if (sv_num<=0) return 0;

// Display all found servers
	printf("\n\n"
	"   | Name           | Dev  | Address         | Ver | Map          | Players "
	"\n");

	for(i=0;i<sv_num;i++) {
		printf("%2d  ",i+1);
		printf("%-16s ",sv_info[i].Name);
		if (sv_device[i]>=NDEV_COM1&&sv_device[i]<=NDEV_COM4) {
			printf("[COM%1d] ",sv_device[i]-NDEV_COM1+1);
			printf("                   ");
		}
		else {
			printf("[IPX]  ");
			printf("%02X:%02X:%02X:%02X:%02X:%02X  ",
				sv_address[i][0],sv_address[i][1],
				sv_address[i][2],sv_address[i][3],
				sv_address[i][4],sv_address[i][5]);
		}
		printf("%d.%d  ",sv_info[i].Version>>8,
			sv_info[i].Version&0xFF);
		printf("%-14s ",sv_info[i].MapName);
		printf("%2d/%2d\n",sv_info[i].NumPlayers,
			sv_info[i].MaxPlayers);
	}

	printf("\nEnter server number or ESC to quit: ");
	fflush(stdout);

	ch=0; idx=0;
	while(ch!=27&&ch!=13) {
		ch=getch();
		if (ch>='0'&&ch<='9') {
			idx=idx*10+(ch-'0');
			putch(ch);
		}
		else if (ch==13) {
			if (idx<1||idx>sv_num||
			    sv_info[idx-1].Version!=NET_VERSION) 
				ch=0;
		}
		else if (ch==8&&idx>0) {
			idx/=10;
			putch(8);putch(32);putch(8);
		}
	}
	printf("\n\n");

	if (ch==27) return 0;

	if (idx>0) idx--;
	*pdevice=sv_device[idx];
	memcpy(paddress,sv_address[idx],6);

	return 1;
}
