/********************************************************************
 EXPLOD.C - Explosion
*********************************************************************/
	
#include "..\engine.h"
#include "..\render.h"
#include "..\h_sound.h"
#include "progs.h"

#define FLAG_SOUND		1


static SPR Explod_spr;
static int Explod_sample;
static int Explod_inited=0;


extern void Particles_MakeExplosion(float *c);


void Explod_LoadSprite(void) 
{
	Explod_inited=SPR_Load(&Explod_spr,"sprites/s_explo3.sp2");
	Explod_sample=SND_LoadSample("rockex.wav");
}


int Explod_Init(int obj_idx,char *params)
{
	Object *o=&Objects[obj_idx];

	if (!Explod_inited) {
		Explod_LoadSprite();
		if (!Explod_inited) return 1;
	}


	o->pob.frame=0;
	o->pob.flags=FLAG_SOUND;
	o->next_frame_tick=TotalTicks+5;
	return 0;
}


int Explod_Event(int obj_idx, int event, int param1, int param2) 
{
	switch(event) {
		case PE_UPDATE:
			return(Explod_Update(obj_idx));
	}
	return 0;
}


int  Explod_Update(int obj_idx)
{
	Object *o=&Objects[obj_idx];

// Model has not been loaded
	if (!Explod_inited) return 1;

// Advance
	if (o->next_frame_tick<=TotalTicks) {
		o->next_frame_tick+=5;
		o->pob.frame++;
		if (o->pob.frame>=Explod_spr.NumFrames)
			return 1;	// Remove object
	}

	return 0;
}

void Explod_Process(POB *pob)
{
	float rgb[3]={1.0f,1.0f,0.8f};
	int r;

// Do we need to load stuff
	if (!Explod_inited) Explod_LoadSprite();

// First frame, play explosion sound
	if (pob->flags&FLAG_SOUND) {
		Particles_MakeExplosion(pob->c);
		SND_PlaySample3D(Explod_sample,1.0,pob->c);
	}


// Add dynamic light 
	r=Explod_spr.NumFrames>>1;
	r=r-abs(pob->frame-r)+2;
	r*=20;

	render->addDLight(pob->c,rgb,1.0,(float)r);
}

void Explod_Render(POB *pob)
{
//	if (Explod_inited) render->drawSPR(&Explod_spr,pob);
}

void Explod_Free(void) 
{
	if (Explod_inited) {
		SPR_Free(&Explod_spr);
		SND_FreeSample(Explod_sample);
		Explod_inited=0;
	}
}

/********************************************************************
			REGISTRATION
*********************************************************************/

void Explosion_Register(Prog *p)
{
	strcpy(p->prog_name,"explode");
	p->flags=	PF_TYPE_SPRITE;
	p->init=	Explod_Init;
	p->event=	Explod_Event;
	p->process=	Explod_Process;
	p->render=	Explod_Render;
	p->free=	Explod_Free;
}
