/********************************************************************
 EXTCAM.C - External camera
*********************************************************************/

	
#include "..\engine.h"
#include "progs.h"

typedef struct {
	short follow_idx;
} local_vars;



int ExtCamera_Init(int obj_idx,char *params)
{
	Object *o=&Objects[obj_idx];
	local_vars *vars=(local_vars *)o->vars;
	char *val;
	int temp;

	val=Parse_GetKeyValue(params,"follow");
	if (val) sscanf(val,"%d",&temp);
	else temp=-1;
		
	vars->follow_idx=temp;
	Watch_Register(obj_idx,temp);
	return 0;
}



int ExtCamera_Event(int obj_idx, int event, int param1, int param2) 
{
	Object *o=&Objects[obj_idx], *follow;
	local_vars *vars=(local_vars *)o->vars;

	switch(event) {

		case PE_OBJ_WATCH:
			if (vars->follow_idx==param1)
				vars->follow_idx=param2;
			break;
		case PE_UPDATE:
			return(ExtCamera_Update(obj_idx));
	}

	return 0;
}

int  ExtCamera_Update(int obj_idx)
{
	Object *o=&Objects[obj_idx], *follow;
	local_vars *vars=(local_vars *)o->vars;
	FMatrix M;
	float delta[3]={0,-30,20}, d[3];

	if (vars->follow_idx<0) return 1;

	follow=&Objects[vars->follow_idx];
	FMatrix_Identity(M);
	FMatrix_PreZRot(M,follow->pob.angle);
	FMatrix_MulFVec(M,delta,d);
	FVec_AddFVec(follow->pob.c,d,o->pob.c);
	o->pob.angle=follow->pob.angle;

	World_ObjectMoved(obj_idx);

	return 0;
}

void ExtCamera_Process(POB *pob) {}
void ExtCamera_Render(POB *pob) {}
void ExtCamera_Free(void) {}


/********************************************************************
			REGISTRATION
*********************************************************************/

void ExtCamera_Register(Prog *p)
{
	strcpy(p->prog_name,"extcamera");
	p->flags=	PF_TYPE_NONE;
	p->init=	ExtCamera_Init;
	p->event=	ExtCamera_Event;
	p->process=	ExtCamera_Process;
	p->render=	ExtCamera_Render;
	p->free=	ExtCamera_Free;
}
