/********************************************************************
 MISSILE.C - A missile
*********************************************************************/

	
#include "..\engine.h"
#include "..\render.h"
#include "progs.h"


typedef struct {
	short shooter;
	short has_exploded;
} local_vars;


static MDL Missile_mdl;
static int Missile_inited=0;

void Missile_LoadModel(void)
{
	Missile_inited=MDL_Load(&Missile_mdl,"models/objects/rocket/tris.md2");
}


int Missile_Init(int obj_idx,char *params)
{
	Object *o=&Objects[obj_idx];
	local_vars *vars=(local_vars *)o->vars;
	char *val;
	int temp;

	if (!Missile_inited) Missile_LoadModel();
	if (!Missile_inited) return 1;

	val = Parse_GetKeyValue(params,"shooter");
	if (val) sscanf(val,"%d",&temp);
	else temp=-1;

	vars->shooter=temp;
	Watch_Register(obj_idx,temp);
	vars->has_exploded=0;

// Setup bound box
	o->mins[0]=o->mins[1]=o->mins[2]=-5;
	o->maxs[0]=o->maxs[1]=o->maxs[2]=5;

	return 0;
}


int Missile_Event(int obj_idx, int event, int param1, int param2) 
{
	Object *o=&Objects[obj_idx];
	local_vars *vars=(local_vars *)o->vars;

	switch(event) {

		case PE_UPDATE:
			return(Missile_Update(obj_idx));

		case PE_COLLISION:
		// Hurt target
			if (param1>=0) 
				Prog_Event(obj_idx,PE_HURT,param1,2);
		// Mark for removal
			vars->has_exploded=1;

			break;

		case PE_OBJ_WATCH:
			if (vars->shooter==param1)
				vars->shooter=param2;
			break;
	}

	return 0;
}


int  Missile_Update(int obj_idx)
{
	Object *o=&Objects[obj_idx], *ex;
	local_vars *vars=(local_vars *)o->vars;
	float delta[3];

	ApplySpeed(obj_idx,delta);
	ClipMoveProjectile(&WorldBSP,obj_idx,delta,vars->shooter);
	World_ObjectMoved(obj_idx);

// Remove when collided
	if (vars->has_exploded) {
	// Add explosion object
		ex=&Objects[World_AddObject(PROG_EXPLOSION,NULL)];
		FVec_Copy(o->pob.c,ex->pob.c);
		ex->leaf_idx=o->leaf_idx;
	}

	return(vars->has_exploded);
}

void Missile_Process(POB *pob)
{
	float rgb[3]={0.8f,0.8f,0.6f};
	render->addDLight(pob->c,rgb,1.0,150.0);
}

void Missile_Render(POB *pob)
{
	if (!Missile_inited)
		Missile_LoadModel();
	if (Missile_inited)
		render->drawMDL(&Missile_mdl,pob);
}

void Missile_Free(void) 
{
	if (Missile_inited) {
		MDL_Free(&Missile_mdl);
		Missile_inited=0;
	}
}


/********************************************************************
			REGISTRATION
*********************************************************************/

void Missile_Register(Prog *p)
{
	strcpy(p->prog_name,"missile");
	p->flags=	PF_TYPE_MODEL;
	p->init=	Missile_Init;
	p->event=	Missile_Event;
	p->process=	Missile_Process;
	p->render=	Missile_Render;
	p->free=	Missile_Free;
}
