/********************************************************************
 PLAYER.C - The main player PROG
*********************************************************************/

	
#include "..\engine.h"
#include "..\render.h"
#include "..\h_sound.h"
#include "progs.h"

#define FLAG_SOUND_PAIN		1
#define FLAG_SOUND_SHOOT	2


//extern int do_coldet;


extern void Shoot_Bullet(int obj_idx);
extern void Shoot_Fireball(int obj_idx);



typedef struct {
	short health;
	short recharge;
} local_vars;

static MDL Player_mdl;
static int Player_inited=0;
static int Shoot_sample;

static AnimSequence
	player_stand,		// Player standing
	player_run,		// Player running
	player_rockatt,		// Player shooting rockets
	player_pain,		// Player being hurt
	player_die1;		// Player death


void Player_LoadModel(void)
{

	//if (!MDL_Load(&Player_mdl,"models/monsters/soldier/tris.md2"))
	if (!MDL_Load(&Player_mdl,"players/male/tris.md2"))
		return;

	MDL_LoadSkin(&Player_mdl,"players/male/cipher.pcx");

// Setup actions
	Anim_Make(&player_stand,&Player_mdl,"stand");
	Anim_Make(&player_run,&Player_mdl,"run");
	Anim_Make(&player_rockatt,&Player_mdl,"rockatt");
	Anim_Make(&player_pain,&Player_mdl,"pain");
	Anim_Make(&player_die1,&Player_mdl,"deatha");

	Shoot_sample=SND_LoadSample("launcher.wav");
	Player_inited=1;
}

int Player_Init(int obj_idx, char *params)
{
	Object *o=&Objects[obj_idx];
	local_vars *vars=(local_vars *)o->vars;

	if (!Player_inited) Player_LoadModel();


//	Object_Anim_Loop(&player_stand,o);
	vars->health=10;
	vars->recharge=0;

// Setup bound box
	o->mins[0]=o->mins[1]=o->mins[2]=-30;
	o->maxs[0]=o->maxs[1]=o->maxs[2]=30;

	return 0;
}


int Player_Update(int obj_idx)
{
	Object *o=&Objects[obj_idx];
	local_vars *vars=(local_vars *)o->vars;
	float delta[4];
	int *pf,idx, i;

// Recharge the weapon
	if (vars->recharge) {
		vars->recharge-=FrameTicks;
		if (vars->recharge<0)
			vars->recharge=0;
	}

// Apply physics stuff
	ApplyFriction(obj_idx);
	ApplyGravity(obj_idx);

	ApplySpeed(obj_idx,delta);

//	if (do_coldet) {
		if (ClipMovePlayer(&WorldBSP,obj_idx,delta)) {
			o->is_gravity=1;
		}
		else {
			if (o->is_gravity) {
				o->is_gravity=0;
				o->speed[2]=-2;
			}
		}
		World_ObjectMoved(obj_idx);
//	}


// Process animations
	if (!Player_inited) return 0;

	pf=&o->pob.frame;

// Check if dead
/*	if (IsAnimSequence(&player_die1,*pf)) {
		if (Object_Anim_Loop(&player_die1,o)) 
			*pf=player_die1.last;
		idx=GetPlayerFromObject(obj_idx);
		if (idx>=0) Players[idx].body_roll+=PI/16384;
		return 0;
	}


// Process animations 
	if (IsAnimSequence(&player_pain,*pf)) {
		if (!Object_Anim_Loop(&player_pain,o)) return 0;
	}

	if (IsAnimSequence(&player_rockatt,*pf)) {
		if (!Object_Anim_Loop(&player_rockatt,o)) return 0;
	}
*/
	if (fabs(o->speed[0])>1.0||fabs(o->speed[1])>1.0) { // Running
		Object_Anim_Loop(&player_run,o);
	}
	else {			// Standing
		Object_Anim_Loop(&player_stand,o);
	}

	return 0;
}



void Player_Attack(int obj_idx)
{
	Object *o=&Objects[obj_idx];
	local_vars *vars=(local_vars *)o->vars;
	int idx, *pf=&o->pob.frame;

	if (vars->health<=0) return;

//	if (!IsAnimSequence(&player_rockatt,*pf)&&
//	    !IsAnimSequence(&player_pain,*pf)) {
///	if (vars->recharge<=0) {

	// Shooting launcher (11 frames times 10)
		vars->recharge=120;
		o->pob.flags|=FLAG_SOUND_SHOOT;
//		Object_Anim_Loop(&player_rockatt,o);
//		Shoot_Bullet(obj_idx);
		Shoot_Fireball(obj_idx);
		idx=GetPlayerFromObject(obj_idx);
		if (idx>=0)
			Players[idx].flags|=PLAYER_FLAG_START_ATTACK;
///	}
}


void Player_Death(int obj_idx)
{
	Object *o=&Objects[obj_idx];
	local_vars *vars=(local_vars *)o->vars;
	int idx;

	if (vars->health<=0) return;

	vars->health=0;
//	Object_Anim_Loop(&player_die1,o);
	idx=GetPlayerFromObject(obj_idx);
	if (idx>=0) {
		Players[idx].flags|=PLAYER_FLAG_DEAD;
		Players[idx].body_roll=PI/3;
		Players[idx].cam_height=-12;
	}
}


void Player_Hurt(int obj_idx,int shooter,int damage)
{
	Object *o=&Objects[obj_idx];
	local_vars *vars=(local_vars *)o->vars;
	int *pf=&o->pob.frame;

	if (vars->health<=0) return;

	if (vars->health<=damage) {
		Player_Death(obj_idx);
	}
	else  {
		vars->health-=damage;
//		if (!IsAnimSequence(&player_pain,*pf)) {
//			Object_Anim_Loop(&player_pain,o);
			o->pob.flags|=FLAG_SOUND_PAIN;
//		}
	}
}




int Player_Event(int obj_idx, int event, int param1, int param2) 
{
	switch(event) {
		case PE_UPDATE:
			return(Player_Update(obj_idx));
		case PE_HURT:
			Player_Hurt(obj_idx,param1,param2);
			break;
		case PE_ATTACK:
			Player_Attack(obj_idx);
			break;
		case PE_DEATH:
			Player_Death(obj_idx);
			break;
	}


	return 0;
}



void Player_Process(POB *pob)
{
	if (!Player_inited) Player_LoadModel();

//	if (pob->flags&FLAG_SOUND_PAIN)
//		World_AddDSound(pob->c,1.0,SMP_SAMPLES_PAIN1);
	if (pob->flags&FLAG_SOUND_SHOOT)
		SND_PlaySample3D(Shoot_sample,0.6,pob->c);
}

void Player_Render(POB *pob)
{
	if (!Player_inited)
		Player_LoadModel();
	if (Player_inited)
		render->drawMDL(&Player_mdl,pob);
}

void Player_Free(void)
{
	if (Player_inited) {
		MDL_Free(&Player_mdl);
		SND_FreeSample(Shoot_sample);
		Player_inited=0;
	}
}


/********************************************************************
			REGISTRATION
*********************************************************************/

void Player_Register(Prog *p)
{
	strcpy(p->prog_name,"player");
	p->flags=	PF_TYPE_MODEL | PF_COL_CLIP;
	p->init=	Player_Init;
	p->event=	Player_Event;
	p->process=	Player_Process;
	p->render=	Player_Render;
	p->free=	Player_Free;
}
