/********************************************************************
 ROTATE.C - Rotating entity (bsp model)
*********************************************************************/

	
#include "..\engine.h"
#include "..\render.h"
#include "progs.h"



typedef struct {
	float speed;		// Speed of move (per tick)
} local_vars;





int Rotate_Init(int obj_idx,char *params) 
{
	Object *o=&Objects[obj_idx];
	local_vars *vars=(local_vars *)o->vars;
	dmodel_t *bspmodel;
	int model_idx,a=0;
	float delta[3];
	char *val;

// No params - should not happen
	if (!params) return 1;

// Find BSP model idx
	val=Parse_GetKeyValue(params,"model");
	if (!val) return 1;

	sscanf(val+1,"%d",&model_idx);
	if (model_idx<1||model_idx>=WorldBSP.NumModels) return 1;

// Setup bsp model info
	o->pob.frame=model_idx;
	WorldBSP.ModelObjectIndex[model_idx]=obj_idx;
	bspmodel=&WorldBSP.Models[model_idx];

// Get spawnflags
	o->pob.vangle=2;
	val=Parse_GetKeyValue(params,"spawnflags");
	if (val) {
		sscanf(val,"%d",&a);
		if (a&4) o->pob.vangle=0;
		else if (a&8) o->pob.vangle=1;
	}

// Get origin
	val=Parse_GetKeyValue(params,"origin");
	if (val) sscanf(val,"%f %f %f",&delta[0],&delta[1],&delta[2]);
	FVec_AddFVec(bspmodel->origin,delta,bspmodel->origin);
// Setup POB
	FVec_Copy(bspmodel->mins,o->mins);
	FVec_Copy(bspmodel->maxs,o->maxs);
	FVec_Copy(bspmodel->origin,o->pob.c);



	return 0;
}


void Rotate_Update(int obj_idx) 
{
	Object *o=&Objects[obj_idx];
	local_vars *vars=(local_vars *)o->vars;

	o->pob.angle+=0.01f;
}

int Rotate_Event(int obj_idx, int event, int param1, int param2) 
{
	Object *o=&Objects[obj_idx];
	local_vars *vars=(local_vars *)o->vars;

	switch(event) {
		case PE_UPDATE:
			Rotate_Update(obj_idx);
			break;
	}

	return 0;
}



void Rotate_Process(POB *pob) {}

void Rotate_Render(POB *pob)
{
	render->drawBSPModel(pob);
}

void Rotate_Free(void) {}


/********************************************************************
			REGISTRATION
*********************************************************************/

void Rotate_Register(Prog *p)
{
	strcpy(p->prog_name,"func_rotating");
	p->flags=	PF_TYPE_BSP_MODEL | PF_COL_CLIP;
	p->init=	Rotate_Init;
	p->event=	Rotate_Event;
	p->process=	Rotate_Process;
	p->render=	Rotate_Render;
	p->free=	Rotate_Free;
}
