/********************************************************************
 TEST.C - A test model (pickable)
*********************************************************************/

	
#include "..\engine.h"
#include "..\render.h"
#include "..\h_sound.h"
#include "progs.h"

#define FLAG_SOUND		1


typedef struct {
	byte done;
} local_vars;

static MDL Test_mdl;
static int Test_sample;
static int Test_inited=0;


void Test_LoadModel() 
{
	Test_inited=MDL_Load(&Test_mdl,"models/items/armor/jacket/tris.md2");
	Test_sample=SND_LoadSample("pkup.wav");
}


int Test_Init(int obj_idx,char *params)
{
	Object *o=&Objects[obj_idx];
	local_vars *vars=(local_vars *)o->vars;
	char *val;

	if (!Test_inited) {
		Test_LoadModel();
		if (!Test_inited) return 1;
	}

	val=Parse_GetKeyValue(params,"origin");

	if (val) {
		sscanf(val,"%f %f %f",
			&o->pob.c[0],
			&o->pob.c[1],
			&o->pob.c[2]);
	}

	vars->done=0;

// Setup bound box
	o->mins[0]=o->mins[1]=o->mins[2]=-10;
	o->maxs[0]=o->maxs[1]=o->maxs[2]=10;

	return 0;
}


int Test_Event(int obj_idx, int event, int param1, int param2) 
{
	Object *o=&Objects[obj_idx];
	local_vars *vars=(local_vars *)o->vars;

	switch(event) {
		case PE_UPDATE:
			if (vars->done) {
				if (vars->done>1) return 1;
				vars->done++;
				o->pob.flags|=FLAG_SOUND;
			}
			o->pob.angle+=PI/256.0;
			break;
		case PE_COLLISION:
			if (!vars->done) vars->done++;
			break;
	}
	return 0;
}


void Test_Process(POB *pob) 
{
	if (!Test_inited) Test_LoadModel();

	if (pob->flags&FLAG_SOUND)
		SND_PlaySample3D(Test_sample,1.0,pob->c);
}

void Test_Render(POB *pob)
{
	if (Test_inited) render->drawMDL(&Test_mdl,pob);
}

void Test_Free(void) 
{
	if (Test_inited) {
		MDL_Free(&Test_mdl);
		SND_FreeSample(Test_sample);
		Test_inited=0;
	}
}

/********************************************************************
			REGISTRATION
*********************************************************************/

void Test_Register(Prog *p)
{
	strcpy(p->prog_name,"test");
	p->flags=	PF_TYPE_MODEL | PF_COL_CHECK;
	p->init=	Test_Init;
	p->event=	Test_Event;
	p->process=	Test_Process;
	p->render=	Test_Render;
	p->free=	Test_Free;
}
