#include "rogl.h"


// Functions from DDWRAP.CPP
extern BOOL DD_SetVideoMode(HWND hwnd, int width, int height, int colors);
extern void DD_Shut(void);


// DLL function export
static RenderExport rexp;

// Imported data
static HINSTANCE hInstance;
static int *pDone, *pPaused;
byte *TexPalette;
int *pLightMode, *pDebugMode;

// DLL data
static char szAppName[] = "GLPoly";
static HWND hwnd = 0;

// OpenGL stuff
static HDC   hdc = 0;
static HGLRC hrc = 0;


__declspec(dllexport) RenderExport *GetRender(RenderImport *rimp)
{

// Get data from the main app
	hInstance=rimp->hInstance;
	pDone=rimp->Done;
	pPaused=rimp->Paused;
	TexPalette=rimp->Palette;
	pDebugMode=rimp->DebugMode;
	pLightMode=rimp->LightMode;

// Setup function pointers
	rexp.init = GL_Init;
	rexp.shut = GL_Shut;
    rexp.beginFrame = GL_BeginFrame;
    rexp.endFrame = GL_EndFrame;
	rexp.getScrWidth = GL_GetScrWidth;
	rexp.getScrHeight = GL_GetScrHeight;
	rexp.getFlags = GL_GetFlags;
	rexp.addDLight = GL_AddDLight;
	rexp.beginBSP = GL_BeginBSP;
	rexp.drawBSPModel = GL_DrawBSPModel;
	rexp.endBSP = GL_EndBSP;
	rexp.drawMDL = GL_DrawMDL;
	rexp.drawSPR = GL_DrawSPR;
	rexp.drawParticles = GL_DrawParticles;
	return(&rexp);
}


/*******************************************************************
				OGL and Windows initialization
********************************************************************/



void GLWinResize(int w,int h)
{
	glViewport(0, 0, w, h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0, (GLfloat)w/(GLfloat)h, 2.0, 4000.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}


LRESULT CALLBACK GLWndProc(HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam)
{
	PAINTSTRUCT ps;
	LRESULT lRc;

	switch (message) {
		case WM_CREATE:
			break;

		case WM_PAINT:
			BeginPaint(hwnd, &ps);
			EndPaint(hwnd, &ps);
			break;

		case WM_SIZE:
//3dfx		GetClientRect(hwnd, &rect);
//			GLWinResize(rect.right-rect.left,rect.bottom-rect.top);
			break;

		case WM_KEYDOWN:
			switch(wParam) {
				case VK_ESCAPE:
					PostMessage(hwnd, WM_CLOSE, 0, 0);
					break;
			}
			break;

		case WM_CLOSE:
			*pDone=1;
			if (hrc) wglDeleteContext(hrc);
			if (hdc) ReleaseDC(hwnd,hdc);
			hrc=0; hdc=0;
			DestroyWindow(hwnd);
			break;

		case WM_DESTROY:
			*pDone=1;
			if (hrc) wglDeleteContext(hrc);
			if (hdc) ReleaseDC(hwnd,hdc);
			hrc=0; hdc=0;
			PostQuitMessage (0);
			break;

		case WM_ACTIVATE:
			*pPaused = (wParam==WA_INACTIVE);
			PostMessage(hwnd, WM_SYNCACQUIRE, 0, 0L);
			break;

		case WM_ENTERMENULOOP:
		case WM_ENTERSIZEMOVE:
			*pPaused = TRUE;
			PostMessage(hwnd, WM_SYNCACQUIRE, 0, 0L);
			break;

		case WM_EXITMENULOOP:
		case WM_EXITSIZEMOVE:
			*pPaused = GetActiveWindow() != hwnd || IsIconic(hwnd);
			PostMessage(hwnd, WM_SYNCACQUIRE, 0, 0L);
			break;

		case WM_SYNCACQUIRE:
			//DI_SyncAcquire();
			break;

		case WM_SYSCOMMAND:
		// Ignore ALT - menu key
			if (wParam==SC_KEYMENU) return 0;

		// Some weird stuff for mouse
			lRc = DefWindowProc(hwnd, message, wParam, lParam);
			if (IsWindow(hwnd))
				PostMessage(hwnd, WM_SYNCACQUIRE, 0, 0L);
			return lRc;
        }    

        return DefWindowProc(hwnd, message, wParam, lParam);
}


BOOL SetupPixelFormat(HDC hdc)
{
	PIXELFORMATDESCRIPTOR pfd, *ppfd;
	int pixelformat;

	ppfd = &pfd;

	ppfd->nSize = sizeof(PIXELFORMATDESCRIPTOR);
	ppfd->nVersion = 1;
	ppfd->dwFlags = PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
	ppfd->dwLayerMask = PFD_MAIN_PLANE;
	ppfd->iPixelType = PFD_TYPE_RGBA;
	ppfd->cColorBits = 16;
	ppfd->cDepthBits = 16;
	ppfd->cAccumBits = 0;
	ppfd->cStencilBits = 0;

	pixelformat = ChoosePixelFormat(hdc, ppfd);

	if ((pixelformat = ChoosePixelFormat(hdc, ppfd)) == 0) {
		MessageBox(NULL, "ChoosePixelFormat failed", "Error", MB_OK);
		return FALSE;
	}

	if (pfd.dwFlags & PFD_NEED_PALETTE) {
		MessageBox(NULL, "Needs palette", "Error", MB_OK);
		return FALSE;
	}

	if (SetPixelFormat(hdc, pixelformat, ppfd) == FALSE) {
		MessageBox(NULL, "SetPixelFormat failed", "Error", MB_OK);
		return FALSE;
	}

	return TRUE;
}


BOOL GLWinInit(int width, int height, int flags)
{
	WNDCLASS wc;

	wc.style         = 0;
	wc.lpfnWndProc   = (WNDPROC)GLWndProc;
	wc.cbClsExtra    = 0;
	wc.cbWndExtra    = 0;
	wc.hInstance     = hInstance;
	wc.hIcon         = LoadIcon (hInstance, szAppName);
	wc.hCursor       = LoadCursor (NULL,IDC_ARROW);
	wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
	wc.lpszMenuName  = szAppName;
	wc.lpszClassName = szAppName;

	if (!RegisterClass(&wc)) return FALSE;


	if (flags&R_WINDOWED) {
	// Normal window
		hwnd = CreateWindow(szAppName, "Poly/OpenGL", 
			WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_VISIBLE|WS_OVERLAPPEDWINDOW,
			CW_USEDEFAULT,CW_USEDEFAULT,
			width, height,
			NULL, NULL, hInstance, NULL);
	}
	else {
	// Dummy fullscreen window
		hwnd = CreateWindow(szAppName, "Poly/OpenGL", 
			WS_POPUP | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
			0, 0,
			width, height,
			NULL, NULL, hInstance, NULL);
	}

	if (!hwnd) return FALSE;


	if (!(flags&R_WINDOWED)) {
		SetCursor(0);
		ShowCursor(0);
	}

	ShowWindow(hwnd, SW_SHOWNORMAL);
	UpdateWindow(hwnd);


// Use DD to change display mode for fullscreen
	if (!(flags&R_WINDOWED)) {
		if (!DD_SetVideoMode(hwnd,width,height,16)) goto error;
	}

// Setup GL context
	hdc = GetDC(hwnd);
	if (!SetupPixelFormat(hdc)) goto error;

	hrc = wglCreateContext(hdc);
	wglMakeCurrent(hdc,hrc);

	return TRUE;

error:

	//printf("Error in GL init %d\n",ddrval);

	GL_Shut();
	DestroyWindow(hwnd);
	return FALSE;
}




HWND GL_Init(int width, int height, int flags)
{
	RECT rect;
	float point_att[3]={0.01f,0.0f,0.01f};

	if (!GLWinInit(width, height, flags)) return NULL;

	InitGLExtensions();

	GetClientRect(hwnd, &rect);
	GLWinResize(rect.right-rect.left,rect.bottom-rect.top);

	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClearDepth(1.0f);
	glDepthFunc(GL_LEQUAL);
	glShadeModel(GL_SMOOTH);
	glCullFace(GL_FRONT);		// we have opposite winding

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_CULL_FACE);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

// Setup textures
	InitGLTextures();

// Setup point parameters
	if (flag_point_parameters) {
		glEnable(GL_POINT_SMOOTH);
		glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
		glPointSize(40.0f);
		roglPointParameterfEXT(GL_POINT_SIZE_MIN_EXT,2.0f);
		roglPointParameterfEXT(GL_POINT_SIZE_MAX_EXT,40.0f);
		roglPointParameterfvEXT(GL_DISTANCE_ATTENUATION_EXT,point_att);
	}

	return hwnd;
}


void GL_Shut(void)
{
	DD_Shut();
}




Camera *ccam;


void GL_BeginFrame(Camera *C)
{
	ccam=C;

	glClear(GL_DEPTH_BUFFER_BIT);

// Setup camera xform
	glMatrixMode(GL_MODELVIEW);

	glLoadIdentity();

	glRotatef(-180.0f-C->pitch*180.0f/PI, 1,0,0);
	glRotatef(-C->yaw*180.0f/PI,   0,0,1);

	glTranslatef(-C->c[0],-C->c[1],-C->c[2]);
}

void GL_EndFrame()
{
	SwapBuffers(hdc);
	NumVisDLights=0;
}


/*
 General info routines
*/

int GL_GetScrWidth(void) 
{
	RECT rect;
	GetClientRect(hwnd, &rect);
	return(rect.right-rect.left);
}

int GL_GetScrHeight(void) 
{
	RECT rect;
	GetClientRect(hwnd, &rect);
	return(rect.bottom-rect.top);
}




int GL_GetFlags(void) {return (R_TEXTURES_POW2|R_TEXTURES_32BIT);}


void GL_DrawParticles(Particle *p, int type) 
{
	float v[3];

	glDisable(GL_TEXTURE_2D);

	if (type==PARTICLE_LINE) {
		glBegin(GL_LINES);
		while(p) {
			glColor3ubv(&p->ramp[p->ramp_frame<<2]);
			glVertex3fv(p->c);
			FVec_Copy(p->c,v);
			v[2]-=2;
			glVertex3fv(v);
			p=p->next;
		}
	}
	else {

		if (flag_point_parameters) {
			glEnable(GL_BLEND);
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		}

		glBegin(GL_POINTS);
		while(p) {
			glColor3ubv(&p->ramp[p->ramp_frame<<2]);
			glVertex3fv(p->c);
			p=p->next;
		}

		if (flag_point_parameters)
			glDisable(GL_BLEND);
	}

	glEnd();
	glEnable(GL_TEXTURE_2D);
}



