/********************************************************************
 R_MDL.C - Render Model
*********************************************************************/


#include "rogl.h"



typedef struct {
	float x, y, z;
	float r, g, b;
} TriVert;


static float vertnormals[162][3]={
#include "..\anorms.h"
};

static TriVert TriVerts[2048];

static float DirLight[3];
static float AmbLight;


void MakeLights(POB *pob,float *light,float amb)
{
	FMatrix M;

	AmbLight=amb;
	FMatrix_Identity(M);
	FMatrix_PreZRot(M,pob->angle);
	FMatrix_InvMulFVec(M,light,DirLight);
}

float GetLight(int idx)
{
	float dot;

	dot=FVec_DotProduct(DirLight,vertnormals[idx])*.8;
	if (dot<AmbLight) dot=AmbLight;
	return(dot);
}


void MakeFrame(MDL *mdl,POB *pob)
{
	float *translate, *scale;
	daliasframe_t *frame;
	dtrivertx_t *tv;
	TriVert *sv;
	int i, j, num;

	num=mdl->NumVerts;
	frame=mdl->Frames[pob->frame];

	translate=frame->translate;
	scale=frame->scale;

	tv=frame->verts;
	sv=TriVerts;
	for(i=0;i<num;i++,tv++,sv++) {
		sv->x=translate[0]+scale[0]*(float)tv->v[0];
		sv->y=translate[1]+scale[1]*(float)tv->v[1];
		sv->z=translate[2]+scale[2]*(float)tv->v[2];
		sv->r=GetLight((int)tv->lightnormalindex);
		sv->b=sv->g=sv->r;
	}
}


typedef struct {
	float s, t;
	int vert_idx;
} GLVert;

void GL_DrawMDL(MDL *mdl, POB *pob)
{
	float light[3]={0,0.7,0.7};
	TriVert *v;
	GLVert *gv;
	byte *glcmd;
	int i,j,num,w,h,a;


// Prepare lights
	MakeLights(pob,light,0.1);

// Make MDL mesh for this frame
	MakeFrame(mdl,pob);


// Model xform
	glPushMatrix();

	glTranslatef(pob->c[0],pob->c[1],pob->c[2]);

	glRotatef(90+pob->angle*180.0/PI,0,0,1);

	glRotatef(-pob->vangle*180.0/PI, 0,1,0);


// Set model's texture
	Tex_SetModelSkin(mdl,0);

	glcmd=mdl->glCommands;

	while(1) {

		num=*((int *)glcmd); glcmd+=sizeof(int);
		gv=(GLVert *)glcmd;

		if (num>0) {
			glBegin(GL_TRIANGLE_STRIP);
		}
		else if (num<0) {
			glBegin(GL_TRIANGLE_FAN);
			num=-num;
		}
		else break;

		for(j=0;j<num;j++,gv++) {
			v=&TriVerts[gv->vert_idx];
			glColor3fv(&v->r);
			glTexCoord2f(gv->s,gv->t);
			glVertex3fv(&v->x);
		}
		glEnd();
		glcmd+=sizeof(GLVert)*num;
	}

	glPopMatrix();
}

