/********************************************************************
 PR_BLIT.C - Polygon Renderer: Span blitters
*********************************************************************/

#include "rsoft.h"


void blit_fill_z(short *zbuf,int count,int z,int dz)
{
	_asm {
		mov esi, zbuf
		mov ecx, count
		mov eax, z 
		mov edx, dz
		test esi, 3
		jz first_not_odd
		mov [esi], ax
		dec ecx
		add esi, 2
		add eax, edx
	first_not_odd:
		mov edi, ecx
		shr ecx, 1
		or ecx, ecx
		jz last_pixel_z
		mov ebx, eax
		add eax, edx
		shl eax, 16
		add eax, ebx
		mov ebx, edx
		shl dx, 1
		shl ebx, 17
		add edx, ebx
	fillz_loop:
		mov [esi], eax
		add esi, 4
		add eax, edx
		dec ecx
		jnz fillz_loop
	last_pixel_z:
		test edi, 1
		jz end_fillz
		mov [esi], ax
	end_fillz:

	}
}

void BlitFlatSurf(void)
{
	ScrSpan *span;
	BYTE *dest;
	short inv_z, inv_dz;
	short *zbuf;
	float c;
	int count, temp;

	inv_dz=(short)(pr_surf->zi*2.0*65536.0);

	span=pr_surf->FirstSpan;
	while(span) {
		temp = pr_buf_offset+span->y*gfxScrWidth+span->x;
		zbuf=ZBuffer+temp;
		dest=gfxBuffer+temp;

	// Blit z span
		c=pr_surf->zc+pr_surf->zi*(float)span->x+
		  pr_surf->zj*(float)span->y;
		inv_z= (short)(c*2.0*65536.0);
		blit_fill_z(zbuf,span->count,inv_z,inv_dz);

	// Draw flat span
		count=span->count;
		memset(dest,pr_color,count);

		span=span->Next;
	}
}




void BlitPerfectTexSurf(void)
{
	ScrSpan *span;
	BYTE *dest;
	short inv_z, inv_dz;
	short *zbuf;
	float a,b,c,fx,fy,inv_c;
	int u, v, count, temp;


	inv_dz=(short)(pr_surf->zi*2.0*65536.0);

	span=pr_surf->FirstSpan;
	while(span) {
		temp=pr_buf_offset+span->y*gfxScrWidth+span->x;
		zbuf=ZBuffer+temp;
		dest=gfxBuffer+temp;

		fx=(float)span->x; 
		fy=(float)span->y;

		a=pr_surf->uc+pr_surf->ui*fx+pr_surf->uj*fy;
		b=pr_surf->vc+pr_surf->vi*fx+pr_surf->vj*fy;
		c=pr_surf->zc+pr_surf->zi*fx+pr_surf->zj*fy;

	// Blit z span
		inv_z= (short)(c*2.0*65536.0);
		blit_fill_z(zbuf,span->count,inv_z,inv_dz);

	// Draw textured span
		count=span->count;
		while(count>0) {
			inv_c=1.0/c;
			u=(int)(a*inv_c);
			v=(int)(b*inv_c);

			if (u>=0&&u<256&&v>=0&&v<256)
				*dest=pr_tex[v*pr_tex_width+u];

			count--; dest++;
			a+=pr_surf->ui; b+=pr_surf->vi; c+=pr_surf->zi;
		}
		span=span->Next;
	}
}



#define TEX_SUBDIV2	4
#define TEX_SUBDIV	(1<<TEX_SUBDIV2)


void BlitTexSurf(void)
{
	ScrSpan *span;
	BYTE *dest;
	short inv_z, inv_dz;
	short *zbuf;
	float a,b,c,da,db,dc,fx,fy,inv_c;
	int u, v, du, dv, count;
	int x2, x_end;
	int tu, tv, temp;

	inv_dz=(short)(pr_surf->zi*2.0*65536.0);

	da=pr_surf->ui*(float)TEX_SUBDIV;
	db=pr_surf->vi*(float)TEX_SUBDIV;
	dc=pr_surf->zi*(float)TEX_SUBDIV;


	span=pr_surf->FirstSpan;
	while(span) {
		temp=pr_buf_offset+span->y*gfxScrWidth+span->x;
		zbuf=ZBuffer+temp;
		dest=gfxBuffer+temp;

		x2=span->x; x_end=span->x+span->count;
		fx=(float)x2; 
		fy=(float)span->y;

		a=pr_surf->uc+pr_surf->ui*fx+pr_surf->uj*fy;
		b=pr_surf->vc+pr_surf->vi*fx+pr_surf->vj*fy;
		c=pr_surf->zc+pr_surf->zi*fx+pr_surf->zj*fy;

	// Blit z span
		inv_z= (short)(c*2.0*65536.0);
		blit_fill_z(zbuf,span->count,inv_z,inv_dz);

	// Draw textured span

		inv_c=65536.0/c;
		du=u=(int)(a*inv_c);
		dv=v=(int)(b*inv_c);

		while(x2<x_end) {

			x2+=TEX_SUBDIV;

			if (x2>=x_end) {
				count=x_end-(x2-TEX_SUBDIV);
				x2=x_end;

				fx=(float)(count-1);
				a+=pr_surf->ui*fx;
				b+=pr_surf->vi*fx;
				c+=pr_surf->zi*fx;

				inv_c=65536.0/c;
				du=(int)(a*inv_c);
				dv=(int)(b*inv_c);

				if (count>1) {
					du=(du-u)/(count-1);
					dv=(dv-v)/(count-1);
				}
			}
			else {
				count=TEX_SUBDIV;

				a+=da; b+=db; c+=dc;
				inv_c=65536.0/c;
				du=(int)(a*inv_c);
				dv=(int)(b*inv_c);

				du=(du-u)>>TEX_SUBDIV2;
				dv=(dv-v)>>TEX_SUBDIV2;
			}

			while(count>0) {
				tu=u>>16; tv=v>>16;
				*dest=pr_tex[tv*pr_tex_width+tu];
				count--; dest++;
				u+=du; v+=dv;
			}
		}

		span=span->Next;
	}
}


void BlitSkySurf(void)
{
	ScrSpan *span;
	BYTE *dest;
	short inv_z, inv_dz;
	short *zbuf;
	float a,b,c,da,db,dc,fx,fy,inv_c;
	int u, v, du, dv, count;
	int x2, x_end;
	int tu, tv, temp;

	inv_dz=(short)(pr_surf->zi*2.0*65536.0);

	da=pr_surf->ui*(float)TEX_SUBDIV;
	db=pr_surf->vi*(float)TEX_SUBDIV;
	dc=pr_surf->zi*(float)TEX_SUBDIV;


	span=pr_surf->FirstSpan;
	while(span) {
		temp=pr_buf_offset+span->y*gfxScrWidth+span->x;
		zbuf=ZBuffer+temp;
		dest=gfxBuffer+temp;

		x2=span->x; x_end=span->x+span->count;
		fx=(float)x2; 
		fy=(float)span->y;

		a=pr_surf->uc+pr_surf->ui*fx+pr_surf->uj*fy;
		b=pr_surf->vc+pr_surf->vi*fx+pr_surf->vj*fy;
		c=pr_surf->zc+pr_surf->zi*fx+pr_surf->zj*fy;

	// Blit z span
		inv_z= (short)(c*2.0*65536.0);
		blit_fill_z(zbuf,span->count,inv_z,inv_dz);

	// Draw textured span

		inv_c=65536.0/c;
		du=u=(int)(a*inv_c);
		dv=v=(int)(b*inv_c);

		while(x2<x_end) {

			x2+=TEX_SUBDIV;

			if (x2>=x_end) {
				count=x_end-(x2-TEX_SUBDIV);
				x2=x_end;

				fx=(float)(count-1);
				a+=pr_surf->ui*fx;
				b+=pr_surf->vi*fx;
				c+=pr_surf->zi*fx;

				inv_c=65536.0/c;
				du=(int)(a*inv_c);
				dv=(int)(b*inv_c);

				if (count>1) {
					du=(du-u)/(count-1);
					dv=(dv-v)/(count-1);
				}
			}
			else {
				count=TEX_SUBDIV;

				a+=da; b+=db; c+=dc;
				inv_c=65536.0/c;
				du=(int)(a*inv_c);
				dv=(int)(b*inv_c);

				du=(du-u)>>TEX_SUBDIV2;
				dv=(dv-v)>>TEX_SUBDIV2;
			}

			while(count>0) {
				tu=(u>>16)&pr_mask_u; 
				tv=(v>>16)&pr_mask_v;
				*dest=pr_tex[tv*pr_tex_width+tu];
				count--; dest++;
				u+=du; v+=dv;
			}
		}

		span=span->Next;
	}
}








