/********************************************************************
 PR_POLY.C - Polygon Renderer: Active Edge Table etc
*********************************************************************/

// This is just some timing code
//int RDTSC(void);
//#pragma aux RDTSC = 0x0f 0x31 modify [eax edx] value [eax];


#include "rsoft.h"


extern int do_tex, do_perfect;
extern int do_asm, do_count;


#define VERT_PROJECTED		100000.0
#define EDGE_CULLED		-2


#define MAX_SCR_SURFS		4096
#define MAX_SCR_EDGES		8192
#define MAX_SCR_SPANS		32768


static ScrSpan ScrSpans[MAX_SCR_SPANS];
int NumScrSpans;

static ScrSurf ScrSurfs[MAX_SCR_SURFS];
int NumScrSurfs;
static ScrSurf SurfStack;

typedef struct _ScrEdge {
	int x, dx;
	ScrSurf *start_surf, *end_surf;
	struct _ScrEdge *Prev, *Next;
	struct _ScrEdge *NextRemove;
} ScrEdge;

static ScrEdge ScrEdges[MAX_SCR_EDGES];
int NumScrEdges;

static ScrEdge FirstEdge, LastEdge;

static ScrEdge *EdgesToAdd[1024];
static ScrEdge *EdgesToRemove[1024];


static ScrSurf *CurSurf;
static float SurfDist;
static int NumSurfEdges;

static float ClipZ1[3], ClipZ2[3]; // Projected !!!
static float ClipLeft1, ClipLeft2;
static float ClipRight1, ClipRight2;
static int ClipFlag;

#define CLIP_LEFT1	1
#define CLIP_LEFT2	2
#define CLIP_RIGHT1	4
#define CLIP_RIGHT2	8
#define CLIP_Z1		16
#define CLIP_Z2		32


void ClearEdgesLists(void);
void AddClippedEdges(void);

void AddEdge(float x1, float y1, float x2, float y2, int edge_idx);
void AddVerticalEdge(float x1,float y1,float y2, int is_lead);
void AddEdgeToLists(int x, int dx, int y1, int y2, int is_lead, int edge_idx);

void PR_BeginFrame()
{

	SurfStack.FirstSpan=NULL;
	SurfStack.zc=-100000.0;
	SurfStack.zi=SurfStack.zj=0.0;

	FirstEdge.Prev=NULL;
	FirstEdge.x=-1000<<16;
	FirstEdge.dx=0;
	FirstEdge.start_surf=&SurfStack;
	FirstEdge.end_surf=NULL;

	LastEdge.Next=NULL;
	LastEdge.x=ccam->width<<16;
	LastEdge.dx=0;
	LastEdge.start_surf=NULL;
	LastEdge.end_surf=&SurfStack;

	NumScrSurfs=0;
	NumScrEdges=0;
	NumScrSpans=0;
	ClearEdgesLists();
}


void PR_BeginBSPFace(void)
{
	SurfDist=100000.0;
	CurSurf=&ScrSurfs[NumScrSurfs];
	NumSurfEdges=0;
	ClipFlag=0;
}

void PR_BeginSkyFace(void)
{
	PR_BeginBSPFace();
}


void PR_EndSkyFace(int side_idx,float *vec_u,float *vec_v,float *plane_normal)
{
	float plane_dist=SKYBOX_SIZE;
	float n[3];
	float d, inv_d, px, py, t;
	float U[3], V[3], u0,v0;
	int temp;


	AddClippedEdges();

//printf("skyface %d\n",side_idx);

// No edges of this poly added to Edge Table
	if (NumSurfEdges<2) {
		return;
	}

// Out of surfaces?
	if (NumScrSurfs<MAX_SCR_SURFS)
		NumScrSurfs++;
	else {
		Engine_FatalError("Out of surfaces.");
	}

	CurSurf->Next=NULL;
	CurSurf->FirstSpan=NULL;
	CurSurf->state=0;

	CurSurf->face_idx=-(side_idx+1);


// Projection constants
	px=1.0/ccam->proj;
	py=1.0/(ccam->proj*ccam->proj_ratio);

// Xform plane:  world -> view
	FMatrix_InvMulFVec(ccam->ToWorld,plane_normal,n);
	d=10000.0;//plane_dist-FVec_DotProduct(plane_normal,ccam->c);
	inv_d=1.0/d;

// Xform texmap vectors: world -> view
	FMatrix_InvMulFVec(ccam->ToWorld,vec_u,U);
	FMatrix_InvMulFVec(ccam->ToWorld,vec_v,V);

// Texture offset
	u0=cbsp->Sky.width>>1;
	v0=cbsp->Sky.height>>1;


// Setup U/Z, V/Z and 1/Z gradients

	CurSurf->zi=n[0]*px*inv_d;
	CurSurf->zj=n[1]*py*inv_d;
	CurSurf->zc=n[2]*inv_d;

// Note that u0 and v0 are incorporated into U and V

	CurSurf->ui=U[0]*px+u0*CurSurf->zi;
	CurSurf->uj=U[1]*py+u0*CurSurf->zj;
	CurSurf->uc=U[2]   +u0*CurSurf->zc;

	CurSurf->vi=V[0]*px+v0*CurSurf->zi;
	CurSurf->vj=V[1]*py+v0*CurSurf->zj;
	CurSurf->vc=V[2]   +v0*CurSurf->zc;

	CurSurf->zc+=-CurSurf->zi*ccam->mid_x-CurSurf->zj*ccam->mid_y;
	CurSurf->uc+=-CurSurf->ui*ccam->mid_x-CurSurf->uj*ccam->mid_y;
	CurSurf->vc+=-CurSurf->vi*ccam->mid_x-CurSurf->vj*ccam->mid_y;
}


void PR_EndBSPFace(int face_idx)
{
	dface_t *face;
	dplane_t *plane;
	texinfo_t *texinfo;
	CacheFace *info;
	float n[3];
	float d, inv_d, px, py, t;
	float U[3], V[3], u0,v0;
	int temp;

	AddClippedEdges();

// No edges of this poly added to Edge Table
	if (NumSurfEdges<2) {
		return;
	}

// Out of surfaces?
	if (NumScrSurfs<MAX_SCR_SURFS)
		NumScrSurfs++;
	else {
		Engine_FatalError("Out of surfaces.");
	}

	CurSurf->Next=NULL;
	CurSurf->FirstSpan=NULL;
	CurSurf->state=0;

	CurSurf->face_idx=face_idx;

	face=&cbsp->Faces[face_idx];
	plane=&cbsp->Planes[face->planenum];
	texinfo=&cbsp->TexInfo[face->texinfo];
	info=&cbsp->CacheFaces[face_idx];

//	CurSurf->texture=cbsp->Textures[cbsp->TextureList[face->texinfo]].miptex;
	temp=SurfDist;
	if (temp<0) temp=0;
	else {temp>>=8; if (temp>3) temp=3;}
	CurSurf->mip_level=temp;

//CurSurf->mip_level=0;


// Projection constants
	px=1.0/ccam->proj;
	py=1.0/(ccam->proj*ccam->proj_ratio);

// Xform plane:  world -> view
	FMatrix_InvMulFVec(ccam->ToWorld,plane->normal,n);
	d=plane->dist-FVec_DotProduct(plane->normal,ccam->c);
	inv_d=1.0/d;

// Xform texmap vectors: world -> view
	FMatrix_InvMulFVec(ccam->ToWorld,texinfo->vecs[0],U);
	FMatrix_InvMulFVec(ccam->ToWorld,texinfo->vecs[1],V);

// Texture offset
	u0=FVec_DotProduct(ccam->c,texinfo->vecs[0])+texinfo->vecs[0][3]-info->min_s;
	v0=FVec_DotProduct(ccam->c,texinfo->vecs[1])+texinfo->vecs[1][3]-info->min_t;

// Adjust tex stuff according to mip level
	if (CurSurf->mip_level>0) {
		t=1.0/(float)(1<<CurSurf->mip_level);
		u0*=t; v0*=t;
		FVec_MulScalar(U,t,U);
		FVec_MulScalar(V,t,V);
	}

// Setup U/Z, V/Z and 1/Z gradients

	CurSurf->zi=n[0]*px*inv_d;
	CurSurf->zj=n[1]*py*inv_d;
	CurSurf->zc=n[2]*inv_d;

// Note that u0 and v0 are incorporated into U and V

	CurSurf->ui=U[0]*px+u0*CurSurf->zi;
	CurSurf->uj=U[1]*py+u0*CurSurf->zj;
	CurSurf->uc=U[2]   +u0*CurSurf->zc;

	CurSurf->vi=V[0]*px+v0*CurSurf->zi;
	CurSurf->vj=V[1]*py+v0*CurSurf->zj;
	CurSurf->vc=V[2]   +v0*CurSurf->zc;

	CurSurf->zc+=-CurSurf->zi*ccam->mid_x-CurSurf->zj*ccam->mid_y;
	CurSurf->uc+=-CurSurf->ui*ccam->mid_x-CurSurf->uj*ccam->mid_y;
	CurSurf->vc+=-CurSurf->vi*ccam->mid_x-CurSurf->vj*ccam->mid_y;
}


void AddClippedEdges(void)
{
// Add clipping edges
	if (ClipFlag&(CLIP_Z1|CLIP_Z2)) {
		AddEdge(ClipZ1[0],ClipZ1[1],ClipZ2[0],ClipZ2[1],-1);
	}
	if (ClipFlag&(CLIP_LEFT1|CLIP_LEFT2)) {
		if (ClipLeft1<ClipLeft2)
			AddVerticalEdge(
				-0.5,
				ClipLeft1,
				ClipLeft2,1);
	}
	if (ClipFlag&(CLIP_RIGHT1|CLIP_RIGHT2)) {
		if (ClipRight1>ClipRight2)
			AddVerticalEdge(
				(float)ccam->width-0.5,
				ClipRight2,
				ClipRight1,0);
	}
}



void PR_AddBSPEdge(int edge_idx, int v1_idx, int v2_idx)
{
	CacheEdge *einfo;
	CacheVert *v1info, *v2info;
	ScrEdge *e;
	float *p1, *p2;
	float v[3], aa[3], t;
	int idx;

// Check if this edge is already in the Edge Table
	einfo=&cbsp->CacheEdges[edge_idx];
	if (einfo->TimeStamp==TotalFrames) {
		idx=einfo->ScrEdgeIdx;
	// Edge has been culled before
		if (idx==EDGE_CULLED) return;
	// Edge has been processed before (this frame)
		if (idx>=0) {
			e=&ScrEdges[idx];
			if (e->start_surf) e->end_surf=CurSurf;
			else e->start_surf=CurSurf;
			NumSurfEdges++;
		// Hack to calc SurfDist
			p1=cbsp->Verts[v1_idx].point;
			FVec_SubFVec(p1,ccam->c,v);
			t=fabs(v[0]);
			t=max(t,fabs(v[1]));
			t=max(t,fabs(v[2]));
			SurfDist=min(SurfDist,t);
		// Done with this edge
			return;
		}
	}

// This edge has not been processed yet :(
/////////////////////////////////////////////////////

// Try to get cached xformed/projected verts
	v1info=&cbsp->CacheVerts[v1_idx];
	v2info=&cbsp->CacheVerts[v2_idx];

// V1 not cached - do xform & project
	if (v1info->TimeStamp!=TotalFrames) {
		v1info->TimeStamp=TotalFrames;
		XformVertex(ccam,cbsp->Verts[v1_idx].point,v1info->c);
		t=v1info->c[2];
		if (t>=MIN_Z) {
			ProjectVertex(ccam,v1info->c,v1info->c);
			v1info->c[2]=VERT_PROJECTED+t;
		}
	}

// V2 not cached - do xform & project
	if (v2info->TimeStamp!=TotalFrames) {
		v2info->TimeStamp=TotalFrames;
		XformVertex(ccam,cbsp->Verts[v2_idx].point,v2info->c);
		t=v2info->c[2];
		if (t>=MIN_Z) {
			ProjectVertex(ccam,v2info->c,v2info->c);
			v2info->c[2]=VERT_PROJECTED+t;
		}
	}

// Projected coords (hopefully)
	p1=v1info->c;  p2=v2info->c;


// Do Z clipping

// Vert 1 too close
	if (p1[2]<VERT_PROJECTED) {
	// Cull the whole edge
		if (p2[2]<VERT_PROJECTED) {
			einfo->TimeStamp=TotalFrames;
			einfo->ScrEdgeIdx=EDGE_CULLED;
			return;
		}
	// Xform V2 world->camera  (aa = vert 2)
		XformVertex(ccam,cbsp->Verts[v2_idx].point,aa);
	// Make new V1 in camera space
		t=(MIN_Z-p1[2])/(aa[2]-p1[2]);
		v[0]=p1[0]+t*(aa[0]-p1[0]);
		v[1]=p1[1]+t*(aa[1]-p1[1]);
		v[2]=MIN_Z;
		p1=v;
	// Project V1
		ProjectVertex(ccam,p1,p1);
		FVec_Copy(p1,ClipZ2);
		ClipFlag|=CLIP_Z2;
		edge_idx=-1;
		SurfDist=MIN_Z;
	}
// Vert 2 too close
	else if (p2[2]<VERT_PROJECTED) {

	// Xform V1 world->camera  (aa = vert 1)
		XformVertex(ccam,cbsp->Verts[v1_idx].point,aa);
	// Make new V2 in camera space
		t=(MIN_Z-aa[2])/(p2[2]-aa[2]);
		v[0]=aa[0]+t*(p2[0]-aa[0]);
		v[1]=aa[1]+t*(p2[1]-aa[1]);
		v[2]=MIN_Z;
		p2=v;
	// Project V2
		ProjectVertex(ccam,p2,p2);
		FVec_Copy(p2,ClipZ1);
		ClipFlag|=CLIP_Z1;
		edge_idx=-1;
		SurfDist=MIN_Z;
	}
	else {
		SurfDist=min(SurfDist,p1[2]-VERT_PROJECTED);
		SurfDist=min(SurfDist,p2[2]-VERT_PROJECTED);
	}

// Do 2d clipping and add edge to the table
	AddEdge(p1[0],p1[1],p2[0],p2[1],edge_idx);
}



void PR_AddSkyEdge(float *ov1,float *ov2)
{
	ScrEdge *e;
	float v1[3], v2[3];
	float v[3], aa[3], t;
	int idx;


	FMatrix_InvMulFVec(ccam->ToWorld,ov1,v1);
	t=v1[2];
	if (t>=MIN_Z) {
		ProjectVertex(ccam,v1,v1);
		v1[2]=VERT_PROJECTED+t;
	}

	FMatrix_InvMulFVec(ccam->ToWorld,ov2,v2);
	t=v2[2];
	if (t>=MIN_Z) {
		ProjectVertex(ccam,v2,v2);
		v2[2]=VERT_PROJECTED+t;
	}


// Do Z clipping

// Vert 1 too close
	if (v1[2]<VERT_PROJECTED) {
	// Cull the whole edge
		if (v2[2]<VERT_PROJECTED)
			return;

	// Xform V2 world->camera  (aa = vert 2)
		FMatrix_InvMulFVec(ccam->ToWorld,ov2,aa);
	// Make new V1 in camera space
		t=(MIN_Z-v1[2])/(aa[2]-v1[2]);
		v1[0]=v1[0]+t*(aa[0]-v1[0]);
		v1[1]=v1[1]+t*(aa[1]-v1[1]);
		v1[2]=MIN_Z;
	// Project V1
		ProjectVertex(ccam,v1,v1);
		FVec_Copy(v1,ClipZ2);
		ClipFlag|=CLIP_Z2;
	}
// Vert 2 too close
	else if (v2[2]<VERT_PROJECTED) {

	// Xform V1 world->camera  (aa = vert 1)
		FMatrix_InvMulFVec(ccam->ToWorld,ov1,aa);
	// Make new V2 in camera space
		t=(MIN_Z-aa[2])/(v2[2]-aa[2]);
		v2[0]=aa[0]+t*(v2[0]-aa[0]);
		v2[1]=aa[1]+t*(v2[1]-aa[1]);
		v2[2]=MIN_Z;
	// Project V2
		ProjectVertex(ccam,v2,v2);
		FVec_Copy(v2,ClipZ1);
		ClipFlag|=CLIP_Z1;
	}

// Do 2d clipping and add edge to the table
	AddEdge(v1[0],v1[1],v2[0],v2[1],-1);
}



/*
  Edges go counter-clock-wise
*/
void AddEdge(float x1, float y1, float x2, float y2, int edge_idx)
{
	CacheEdge *einfo;
	float slope;
	int top, bot, temp;
	int x, dx, is_lead;
	float xmin=-0.5, xmax=(float)ccam->width-0.5;
	float ymin=-0.5, ymax=(float)ccam->height-0.5;


// Horizontal 2D clipping
	if (x1<x2) {		// Start point 'lefter' than end point
		if (x1<=xmin) {		// Left clip 2
			if (x2<xmin) {
				if (edge_idx>=0) {
					einfo=&cbsp->CacheEdges[edge_idx];
					einfo->TimeStamp=TotalFrames;
					einfo->ScrEdgeIdx=EDGE_CULLED;
				}
				return;
			}
			y1=y1+(y2-y1)*(xmin-x1)/(x2-x1);
			x1=xmin;
			ClipLeft2=y1;
			ClipFlag|=CLIP_LEFT2;
			edge_idx=-1;
		}
		if (x2>=xmax) {		// Right clip 1
			if (x1>xmax) {
				if (edge_idx>=0) {
					einfo=&cbsp->CacheEdges[edge_idx];
					einfo->TimeStamp=TotalFrames;
					einfo->ScrEdgeIdx=EDGE_CULLED;
				}
				return;
			}
			y2=y1+(y2-y1)*(xmax-x1)/(x2-x1);
			x2=xmax;
			ClipRight1=y2;
			ClipFlag|=CLIP_RIGHT1;
			edge_idx=-1;
		}
	}
	else if (x1>x2) {	// End point 'lefter' than start point
		if (x2<=xmin) {		// Left clip 1
			if (x1<xmin) {
				if (edge_idx>=0) {
					einfo=&cbsp->CacheEdges[edge_idx];
					einfo->TimeStamp=TotalFrames;
					einfo->ScrEdgeIdx=EDGE_CULLED;
				}
				return;
			}
			y2=y1+(y2-y1)*(xmin-x1)/(x2-x1);
			x2=xmin;
			ClipLeft1=y2;
			ClipFlag|=CLIP_LEFT1;
			edge_idx=-1;
		}
		if (x1>=xmax) {		// Right clip 2
			if (x2>xmax) {
				if (edge_idx>=0) {
					einfo=&cbsp->CacheEdges[edge_idx];
					einfo->TimeStamp=TotalFrames;
					einfo->ScrEdgeIdx=EDGE_CULLED;
				}
				return;
			}
			y1=y1+(y2-y1)*(xmax-x1)/(x2-x1);
			x1=xmax;
			ClipRight2=y1;
			ClipFlag|=CLIP_RIGHT2;
			edge_idx=-1;
		}
	}
	else {			// Vertical edge
		if (x1<=xmin||x1>=xmax) return;
	}

// Vertical 2D clipping
	if (y1<y2) {		// Start point higher than end
		if (y1<ymin) {
			if (y2<=ymin) return;
			x1=x1+(x2-x1)*(ymin-y1)/(y2-y1);
			y1=ymin;
		}
		if (y2>ymax) {
			if (y1>=ymax) return;
			x2=x1+(x2-x1)*(ymax-y1)/(y2-y1);
			y2=ymax;
		}
	}
	else if (y1>y2) {	// Start point lower than end
		if (y2<ymin) {
			if (y1<=ymin) return;
			x2=x1+(x2-x1)*(ymin-y1)/(y2-y1);
			y2=ymin;
		}
		if (y1>ymax) {
			if (y2>=ymax) return;
			x1=x1+(x2-x1)*(ymax-y1)/(y2-y1);
			y1=ymax;
		}
	}
	else return;

/*
if (	x1<xmin||x2<xmin||x1>xmax||x2>xmax||
	y1<ymin||y2<ymin||y1>ymax||y2>ymax) {

	printf("Error: %f, %f -> %f, %f\n",x1,y1,x2,y2);
	getch();
	exit(1);
}
*/

// Prepare to add the edge

	top=(int)ceil(y1);
	bot=(int)ceil(y2);
	temp=bot-top;
	if (temp==0) return;

// Leading edge (from p1 to p2)
	if (temp>0) {
		is_lead=1;
		slope=(x2-x1)/(y2-y1);
		dx=(int)(slope*65536.0);
		x =(int)(( x1+((float)top-y1)*slope )*65536.0);
	}
// Trailing edge (from p2 to p1)
	else {
		is_lead=0;
		temp=top; top=bot; bot=temp;
		slope=(x1-x2)/(y1-y2);
		dx=(int)(slope*65536.0);
		x =(int)(( x2+((float)top-y2)*slope )*65536.0);
	}

	AddEdgeToLists(x,dx,top,bot,is_lead,edge_idx);
}

void AddVerticalEdge(float x1,float y1,float y2, int is_lead)
{
	int top, bot, temp;
	int x;
	float ymin=-0.5, ymax=(float)ccam->height-0.5;

// y1 is ALWAYS less than y2
	if (y1>=y2) return;

// Vertical 2D clipping
	if (y1<ymin) {
		if (y2<=ymin) return;
		y1=ymin;
	}
	if (y2>ymax) {
		if (y1>=ymax) return;
		y2=ymax;
	}

/*
if (y1<ymin||y2<ymin||y1>ymax||y2>ymax) {
	printf("Error: %f, %f -> %f, %f\n",x1,y1,x1,y2);
	getch();
	exit(1);
}
*/

// Prepare to add the edge

	top=(int)ceil(y1);
	bot=(int)ceil(y2);
	temp=bot-top;
	if (temp<=0) return;

// Find edge start coord (fixed)
	x=(int)(x1*65536.0);

	AddEdgeToLists(x,0,top,bot,is_lead,-1);
}


void AddEdgeToLists(int x, int dx, int y1, int y2, int is_lead, int edge_idx)
{
	CacheEdge *einfo;
	ScrEdge *e;

	if (NumScrEdges>=MAX_SCR_EDGES) {
		Engine_FatalError("Out of edges.");
	}

	if (y1>=y2) return;

//printf("   edge %d %d %d %d\n",x>>16,y1,(x+dx*(y2-y1))>>16,y2);

	if (edge_idx>=0) {
		einfo=&cbsp->CacheEdges[edge_idx];
		einfo->TimeStamp=TotalFrames;
		einfo->ScrEdgeIdx=NumScrEdges;
	}

// Init ScrEdge
	e=&ScrEdges[NumScrEdges];

	e->x=x; 
	e->dx=dx;
	if (is_lead) {
		e->start_surf=CurSurf;
		e->end_surf=NULL;
	}
	else {
		e->start_surf=NULL;
		e->end_surf=CurSurf;
	}

// Add ScrEdge to lists
	e->Next=EdgesToAdd[y1];
	EdgesToAdd[y1]=e;
	e->NextRemove=EdgesToRemove[y2-1];
	EdgesToRemove[y2-1]=e;
	NumScrEdges++;
	NumSurfEdges++;
}


void ClearEdgesLists(void)
{
	ScrEdge **pa, **pe;
	int i;

	pa=EdgesToAdd; pe=EdgesToRemove;
	for(i=0;i<ccam->height+1;i++,pa++,pe++) {
		*pa=NULL; *pe=NULL;
	}
}


void InsertScrEdge(ScrEdge *e)
{
	ScrEdge *pe;

	pe=FirstEdge.Next;
	while(pe->x<e->x) pe=pe->Next;

// Insert 'e' before 'pe'
	e->Next=pe;
	e->Prev=pe->Prev;
	e->Prev->Next=e;
	pe->Prev=e;
}

void RemoveScrEdge(ScrEdge *e)
{
// Remove edge from active list
	e->Prev->Next=e->Next;
	e->Next->Prev=e->Prev;
}

void AddSpan(ScrSurf *surf, int x, int y)
{
	ScrSpan *span;

	if (x<=surf->x_start) return;

	if (NumScrSpans>=MAX_SCR_SPANS) {
		Engine_FatalError("Out of spans.");
	}

	span=&ScrSpans[NumScrSpans];
	NumScrSpans++;

	span->x=surf->x_start; span->y=y;
	span->count=x-surf->x_start;

	span->Next=surf->FirstSpan;
	surf->FirstSpan=span;
}



void ProcessScanLine(int y)
{
	ScrEdge *e, *ne, *pe;
	ScrSurf *surf, *surf2;
	float z, z2, fx, fy;
	int x;


// Sort in new edges into AET
	e=EdgesToAdd[y];
	while(e) {
		ne=e->Next;
		InsertScrEdge(e);
		e=ne;
	}


// Process edges, emit spans
//**************************************
	SurfStack.Next=SurfStack.Prev=&SurfStack;
	SurfStack.x_start=0;
	SurfStack.state=1;

	for(e=FirstEdge.Next;e!=NULL;e=e->Next) {
		x=(e->x+0xFFFF)>>16;
		fx=(float)x;
		fy=(float)y;

	// Trailing edge (surface on stack)
		if (e->end_surf) {
			surf=e->end_surf;
			if (--surf->state==0) {

			// Emit span if surface on top
				if (SurfStack.Next==surf) {
					AddSpan(surf,x,y);
					surf->Next->x_start=x;
				}
			// Remove surface from stack
				surf->Next->Prev=surf->Prev;
				surf->Prev->Next=surf->Next;
			}
		}

	// Leading edge (surface is not on stack yet)
		if (e->start_surf) {
			surf=e->start_surf;
			if (++surf->state==1) {

				z=surf->zc+fx*surf->zi+fy*surf->zj;
				surf2=SurfStack.Next;
				z2=surf2->zc+fx*surf2->zi+fy*surf2->zj;

			// New top surface (emit old span, start a new one)
				if (z>z2) {
					AddSpan(surf2,x,y);
					surf->x_start=x;

					surf->Next=surf2;
					surf2->Prev=surf;
					SurfStack.Next=surf;
					surf->Prev=&SurfStack;
				}
			// Sort the surface into stack
				else {
					do {
						surf2=surf2->Next;
						z2=surf2->zc+fx*surf2->zi
							+fy*surf2->zj;
					} while(z<z2);

				// Insert before surf2
					surf->Next=surf2;
					surf->Prev=surf2->Prev;
					surf2->Prev->Next=surf;
					surf2->Prev=surf;
				}
			}
		}
	}

// Remove edges (next line)
	e=EdgesToRemove[y];
	while(e) {
		RemoveScrEdge(e);
		e=e->NextRemove;
	}

// Advance edges
	e=FirstEdge.Next;
	while(e!=&LastEdge) {
		ne=e->Next;

	// Advance edge x
		e->x+=e->dx;

	// Move the edge back to the proper sorted location
		while (e->x<e->Prev->x) {
			pe=e->Prev;
			pe->Next=e->Next;
			e->Next->Prev=pe;
			pe->Prev->Next=e;
			e->Prev=pe->Prev;
			e->Next=pe;
			pe->Prev=e;
		}

	// Next edge
		e=ne;
	}
}




/**************************************************************
 Global variables to pass parameters to blitter
*/

ScrSurf *pr_surf;	// Surface to blit
int pr_cam_width;	// Camera width (output)
int pr_buf_offset;	// Offset in the output buffer
byte *pr_tex;		// Texture (from surface cache)
int pr_tex_width;	// Texture width
int pr_mask_u,pr_mask_v;// Texture size masks (for sky, water etc)
int pr_color;		// For flat blitters

/*************************************************************/


void PR_EndFrame(void)
{
	ScrSurf *surf;
	int i,t,count,type;

// Process all scan-lines, Emit spans
	FirstEdge.Next=&LastEdge;
	LastEdge.Prev=&FirstEdge;

	for(i=0;i<ccam->height;i++)
		ProcessScanLine(i);

/*
if (FirstEdge.Next!=&LastEdge||LastEdge.Prev!=&FirstEdge) {
	printf("Edges left in AET!!!\n");
	getch();
}
*/

// Blit spans
	pr_cam_width=ccam->width;
	pr_buf_offset=ccam->sy*gfxScrWidth+ccam->sx;

//	count=RDTSC();

	if (do_tex) {
		if (do_perfect) {
			surf=ScrSurfs;
			for(i=0;i<NumScrSurfs;i++,surf++) {
				if (!surf->FirstSpan) continue;
				pr_surf=surf;
				type=GetSurf(
					surf->face_idx,
					surf->mip_level,
					&pr_tex,&pr_tex_width,&t);
				if (type==1)
					BlitPerfectTexSurf();
				else if (type==2) {
					pr_mask_u=pr_tex_width-1;
					pr_mask_v=t-1;
					BlitSkySurf();
				}
				else if (type==0) {
					pr_color=16+(surf->face_idx&15);
					BlitFlatSurf();
				}
			}
		}
		else {
			if (do_asm) {

				surf=ScrSurfs;
				for(i=0;i<NumScrSurfs;i++,surf++) {
					if (!surf->FirstSpan) continue;
					pr_surf=surf;
					type=GetSurf(
						surf->face_idx,
						surf->mip_level,
						&pr_tex,&pr_tex_width,&t);
					if (type==1)
						BlitTex16Surf_asm();
					else if (type==2) {
						pr_mask_u=pr_tex_width-1;
						pr_mask_v=t-1;
						BlitSkySurf();
					}
					else if (type==0) {
						pr_color=16+(surf->face_idx&15);
						BlitFlatSurf();
					}
				}
			}
			else {

				surf=ScrSurfs;
				for(i=0;i<NumScrSurfs;i++,surf++) {
					if (!surf->FirstSpan) continue;
					pr_surf=surf;
					type=GetSurf(
						surf->face_idx,
						surf->mip_level,
						&pr_tex,&pr_tex_width,&t);
					if (type==1)
						BlitTexSurf();
					else if (type==2) {
						pr_mask_u=pr_tex_width-1;
						pr_mask_v=t-1;
						BlitSkySurf();
					}
					else if (type==0) {
						pr_color=16+(surf->face_idx&15);
						BlitFlatSurf();
					}
				}

			}
		}
	}
	else {

		if (do_asm) {
			surf=ScrSurfs;
			for(i=0;i<NumScrSurfs;i++,surf++) {
				if (!surf->FirstSpan) continue;
				pr_surf=surf;
				pr_color=16+(surf->face_idx&15);
				BlitFlatSurf_asm();
			}
		}
		else {
			surf=ScrSurfs;
			for(i=0;i<NumScrSurfs;i++,surf++) {
				if (!surf->FirstSpan) continue;
				pr_surf=surf;
				pr_color=16+(surf->face_idx&15);
				BlitFlatSurf();
			}
		}

	}

/*	count=RDTSC()-count;
	if (do_count) {
		do_count=0;
		printf("Blit (%s): %d\n",do_asm?"a":"c",count>>10);
		getch();
	}
*/
}

