/********************************************************************
 R_MDL.C - Render Model
*********************************************************************/


#include "rsoft.h"


#define CLIP_LEFT	1
#define CLIP_RIGHT	2
#define CLIP_TOP	4
#define CLIP_BOT	8
#define NOCLIP_Z	16

#define TRI_MIN_Z	1.0


static TriVert TriVerts[2048];


static float vertnormals[162][3]={
#include "..\anorms.h"
};

static float DirLight[3];
static float AmbLight;


void MakeLights(POB *pob,float *light,float amb)
{
	FMatrix M;

	AmbLight=amb;
	FMatrix_Identity(M);
	FMatrix_PreZRot(M,pob->angle);
	FMatrix_InvMulFVec(M,light,DirLight);
}

int GetLight(int idx)
{
	float dot;
	int i;

	dot=FVec_DotProduct(DirLight,vertnormals[idx])*.8;
	if (dot<AmbLight) dot=AmbLight;
//	if (dot>1.0) dot=1.0;
	i=(1.0-dot)*(PalNumShades-1)*65536.0;
	return(i);
}


void MakeFrame(MDL *mdl,POB *pob)
{
	FMatrix M,xform;
	float disp[3];
	float t1[3], t2[3], *translate, *scale;
	daliasframe_t *frame;
	dtrivertx_t *tv;
	TriVert *sv;
	int i, j, num;

// Make obj->view xform
	FMatrix_Identity(M);
	FMatrix_PreZRot(M,pob->angle+PI/2);
	FMatrix_PreYRot(M,pob->vangle);
	FMatrix_InvMulFMatrix(ccam->ToWorld,M,xform);
	FVec_SubFVec(pob->c,ccam->c,t1);
	FMatrix_InvMulFVec(ccam->ToWorld,t1,disp);


	num=mdl->NumVerts;
	frame=mdl->Frames[pob->frame];

	translate=frame->translate;
	scale=frame->scale;

	tv=frame->verts;
	sv=TriVerts;
	for(i=0;i<num;i++,tv++,sv++) {
		for(j=0;j<3;j++)
			t1[j]=translate[j]+scale[j]*(float)tv->v[j];

		FMatrix_MulFVec(xform,t1,t2);
		FVec_AddFVec(t2,disp,t1);

		if (t1[2]>=TRI_MIN_Z) {
			sv->clip=NOCLIP_Z;
			ProjectVertex(ccam,t1,t2);

			sv->x=t2[0];
			if (sv->x>ccam->width) sv->clip|=CLIP_RIGHT;
			else if (sv->x<0) sv->clip|=CLIP_LEFT;

			sv->y=t2[1];
			if (sv->y>ccam->height) sv->clip|=CLIP_BOT;
			else if (sv->y<0) sv->clip|=CLIP_TOP;

			sv->x+=ccam->sx;
			sv->y+=ccam->sy;

			sv->inv_z=t2[2]*131072.0; //65536.0*256.0;
			sv->s=GetLight((int)tv->lightnormalindex);
		}
		else
			sv->clip=0;
	}
}




void Render_MDL(MDL *mdl, POB *pob)
{
	float light[3]={0,0.7,0.7};
	TriVert *v1, *v2, *v3;
	dstvert_t *tv1, *tv2, *tv3;
	dtriangle_t *tri;
	byte *skin;
	int x1,y1,x2,y2;
//	float t[3], dist;
	int i,num,skin_width;


// Transform object's world coords into view
//	XformVertex(C,pob->c,t);

// Do 3D trivial object reject (check against frustrum and min_z)
//	dist=mdl->Header->boundingradius;
//	if (t[2]<10.0) return;
//	if (t[2]< t[0]-dist||t[2]< t[1]-dist) return;
//	if (t[2]<-t[0]-dist||t[2]<-t[1]-dist) return;


// Prepare lights
	MakeLights(pob,light,0.1);

// Make projected MDL mesh for this frame
	MakeFrame(mdl,pob);

// Get skin texture
	skin=mdl->Skins[0];
	skin_width=mdl->Header->skinwidth;

// Init Tri renderer
	Tri_SetClip(ccam->sx,ccam->sy,
		ccam->sx+ccam->width,ccam->sy+ccam->height);
	Tri_SetTexture(skin,skin_width);
	Tri_SetColor(15);
	Tri_SetPalTable(PalTables[0]);


	num=mdl->NumTris;
	tri=mdl->Tris;

	for(i=0;i<num;i++,tri++) {
		v1=&TriVerts[tri->index_xyz[0]];
		v2=&TriVerts[tri->index_xyz[1]];
		v3=&TriVerts[tri->index_xyz[2]];

		if ((v1->clip&v2->clip&v3->clip)!=NOCLIP_Z) continue;

	// Do back face culling
		x1=v2->x-v1->x;
		x2=v3->x-v1->x;
		y1=v2->y-v1->y;
		y2=v3->y-v1->y;
		if ((x1*y2-x2*y1)<=0) continue;

	// Setup tex coords for this triangle
		tv1=&mdl->TexVerts[tri->index_st[0]];
		tv2=&mdl->TexVerts[tri->index_st[1]];
		tv3=&mdl->TexVerts[tri->index_st[2]];

		v1->u=tv1->s; v1->v=tv1->t;
		v2->u=tv2->s; v2->v=tv2->t;
		v3->u=tv3->s; v3->v=tv3->t;


	// Draw the triangle
		Tri_Tex_Gouraud_zbuf(v1,v2,v3);
//		Tri_Flat_Gouraud_zbuf(v1,v2,v3);

	}
}
