#include "rsoft.h"


/*int FixMul(int, int);		// Fixed math (386 asm)
#pragma aux FixMul =		\
	"imul ebx"		\
	"shrd eax, edx, 16"	\
	parm [eax] [ebx]	\
	value [eax]		\
	modify [edx];
*/

int FixMul(int a, int b) 
{
	_asm {
		xor edx, edx
		mov eax, a
		mov ebx, b
		imul ebx
		shrd eax, edx, 16
		mov a, eax
	}
	return(a);
}

#define Inverse(a)		(((a)<1024)?(InvTable[a]):(65536/(a)))



typedef struct {
	int x, dx;
	int u, v, z, du, dv, dz;
	int s, ds;
} TriEdge;

static byte *tri_texture;
static int tri_tex_width;
static byte *tri_pal_table;
static int tri_color;
static int tri_clip_x1, tri_clip_y1, tri_clip_x2, tri_clip_y2;

//void blit_fill_z_new(short *zbuf,int count,int z,int dz);



/********************************************************************
			SETUP FOR TRIANGLE RENDERER
*********************************************************************/
static int InvTable[1024];

void Tri_Init() 
{
	int i;
	InvTable[0]=0;
	for(i=1;i<1024;i++)
		InvTable[i]=65536/i;
}


void Tri_SetTexture(byte *texture, int tex_width)
{
	tri_texture=texture;
	tri_tex_width=tex_width;
}

void Tri_SetColor(int color)
{
	tri_color=color;
}

void Tri_SetPalTable(byte *ptr)
{
	tri_pal_table=ptr;
}

void Tri_SetClip(int x1, int y1, int x2, int y2)
{
	tri_clip_x1=x1;
	tri_clip_y1=y1;
	tri_clip_x2=x2;
	tri_clip_y2=y2;
}




/********************************************************************
			DRAW FLAT TRIANGLE
*********************************************************************/

int Tri_EdgeFlat(TriEdge *e,TriVert *v1,TriVert *v2);
void Tri_BlitFlat(TriEdge *e1,TriEdge *e2,int y,int h);

void Tri_Flat(TriVert *v1,TriVert *v2,TriVert *v3)
{
	TriVert *p1,*p2,*p3;
	TriEdge eshort,elong;
	int y1,y2,y3,y,h;

	y1=v1->y; y2=v2->y; y3=v3->y;

	if (y1<y2) {
		if (y1<y3) {
			p1=v1;
			if (y2<y3) {
				p2=v2;
				p3=v3;
			}
			else {	// y2>y3
				p2=v3;
				p3=v2;
			}
		}
		else {	// y3<y1
			p1=v3;
			p2=v1;
			p3=v2;
		}
	}
	else { 	// y1>y2
		if (y2<y3) {
			p1=v2;
			if (y3<y1) {
				p2=v3;
				p3=v1;
			}
			else {	// y3>y1
				p2=v1;
				p3=v3;
			}
		}
		else {	// y2>y3
			p1=v3;
			p2=v2;
			p3=v1;
		}
	}

// Long side
	if (Tri_EdgeFlat(&elong,p1,p3)<=0) return;

// Start y
	y=max(p1->y,tri_clip_y1);

// Top short side
	h=Tri_EdgeFlat(&eshort,p1,p2);
	if (h>0) {
		if (eshort.dx<elong.dx)
			Tri_BlitFlat(&eshort,&elong,y,h);
		else
			Tri_BlitFlat(&elong,&eshort,y,h);
		y+=h;
	}
// Bot short side
	h=Tri_EdgeFlat(&eshort,p2,p3);
	if (h>0) {
		if (eshort.x<elong.x)
			Tri_BlitFlat(&eshort,&elong,y,h);
		else
			Tri_BlitFlat(&elong,&eshort,y,h);
	}
}


int Tri_EdgeFlat(TriEdge *e,TriVert *v1,TriVert *v2)
{
	int dy, h, adv_y=0;

	dy=h=v2->y-v1->y;

	if (v1->y<tri_clip_y1) {
		if (v2->y<=tri_clip_y1) return 0;
		adv_y=tri_clip_y1-v1->y;
		h-=adv_y;
	}
	if (v2->y>tri_clip_y2) {
		if (v1->y>=tri_clip_y2) return 0;
		h-=v2->y-tri_clip_y2;
	}

	dy=Inverse(dy);
	e->x=v1->x<<16;
	e->dx=(v2->x-v1->x)*dy;

	if (adv_y) {
		e->x+=e->dx*adv_y;
	}

	return h;
}

void Tri_BlitFlat(TriEdge *e1,TriEdge *e2,int y,int h)
{
	byte *buf_line, *buf;
	int x1, x2, count;
	DWORD color4;

	buf_line=gfxBuffer+y*gfxScrWidth;
	color4=tri_color+(tri_color<<8)+(tri_color<<16)+(tri_color<<24);

	while(h>0) {

		x1=e1->x>>16; x2=e2->x>>16;
		if (x2>tri_clip_x2) {
			x2=tri_clip_x2;
		}
		if (x1<tri_clip_x1) {
			x1=tri_clip_x1;
		}
		if (x1>=x2) goto NEXT_SPAN;

		buf=buf_line+x1;
		count=(x2-x1)>>2;
		while(count>0) {
			count--;
			*((DWORD *)buf)=color4;
			buf+=4;
		}
		count=(x2-x1)&3;
		while(count>0) {
			count--;
			*buf=tri_color;
			buf++;
		}

	NEXT_SPAN:
		e1->x+=e1->dx; e2->x+=e2->dx;
		buf_line+=gfxScrWidth;
		h--;
	}
}





/********************************************************************
			DRAW FLAT TRIANGLE (ZBUFFER)
*********************************************************************/

int Tri_EdgeFlat_z(TriEdge *e,TriVert *v1,TriVert *v2);
void Tri_BlitFlat_zbuf(TriEdge *e1,TriEdge *e2,int y,int h);


void Tri_Flat_zbuf(TriVert *v1,TriVert *v2,TriVert *v3)
{
	TriVert *p1,*p2,*p3;
	TriEdge eshort,elong;
	int y1,y2,y3,y,h;

	y1=v1->y; y2=v2->y; y3=v3->y;

	if (y1<y2) {
		if (y1<y3) {
			p1=v1;
			if (y2<y3) {
				p2=v2;
				p3=v3;
			}
			else {	// y2>y3
				p2=v3;
				p3=v2;
			}
		}
		else {	// y3<y1
			p1=v3;
			p2=v1;
			p3=v2;
		}
	}
	else { 	// y1>y2
		if (y2<y3) {
			p1=v2;
			if (y3<y1) {
				p2=v3;
				p3=v1;
			}
			else {	// y3>y1
				p2=v1;
				p3=v3;
			}
		}
		else {	// y2>y3
			p1=v3;
			p2=v2;
			p3=v1;
		}
	}

// Long side
	if (Tri_EdgeFlat_z(&elong,p1,p3)<=0) return;

// Start y
	y=max(p1->y,tri_clip_y1);

// Top short side
	h=Tri_EdgeFlat_z(&eshort,p1,p2);
	if (h>0) {
		if (eshort.dx<elong.dx)
			Tri_BlitFlat_zbuf(&eshort,&elong,y,h);
		else
			Tri_BlitFlat_zbuf(&elong,&eshort,y,h);
		y+=h;
	}
// Bot short side
	h=Tri_EdgeFlat_z(&eshort,p2,p3);
	if (h>0) {
		if (eshort.x<elong.x)
			Tri_BlitFlat_zbuf(&eshort,&elong,y,h);
		else
			Tri_BlitFlat_zbuf(&elong,&eshort,y,h);
	}
}


int Tri_EdgeFlat_z(TriEdge *e,TriVert *v1,TriVert *v2)
{
	int dy, h, adv_y=0;

	dy=h=v2->y-v1->y;

	if (v1->y<tri_clip_y1) {
		if (v2->y<=tri_clip_y1) return 0;
		adv_y=tri_clip_y1-v1->y;
		h-=adv_y;
	}
	if (v2->y>tri_clip_y2) {
		if (v1->y>=tri_clip_y2) return 0;
		h-=v2->y-tri_clip_y2;
	}

	dy=Inverse(dy);
	e->x=v1->x<<16;
	e->z=v1->inv_z;
	e->dx=(v2->x-v1->x)*dy;
	e->dz=FixMul(v2->inv_z-v1->inv_z,dy);


	if (adv_y) {
		e->x+=e->dx*adv_y;
		e->z+=e->dz*adv_y;
	}

	return h;
}

void Tri_BlitFlat_zbuf(TriEdge *e1,TriEdge *e2,int y,int h)
{
	byte *buf_line, *buf;
	short *zbuf_line, *zbuf, z, dz;
	int x1, x2, dx, count, x_adv;

	buf_line=gfxBuffer+y*gfxScrWidth;
	zbuf_line=ZBuffer+y*gfxScrWidth;

	while(h>0) {

		x1=e1->x>>16; x2=e2->x>>16;
		dx=x2-x1;
		if (x2>tri_clip_x2) {
			x2=tri_clip_x2;
		}
		if (x1<tri_clip_x1) {
			x_adv=tri_clip_x1-x1;
			x1=tri_clip_x1;
		}
		else x_adv=0;

		if (x1>=x2) goto NEXT_SPAN;

		dx=Inverse(dx);
		dz=FixMul(e2->z-e1->z,dx);
		z=e1->z;

		if (x_adv) {
			z=e1->z+dz*x_adv;
		}

		buf=buf_line+x1;
		zbuf=zbuf_line+x1;
		count=x2-x1;
		while(count>0) {
			if ((*zbuf)<z) {
				*zbuf=z;
				*buf=tri_color;
			}
			count--;
			buf++; zbuf++; z+=dz;
		}

	NEXT_SPAN:
		e1->x+=e1->dx; 
		e1->z+=e1->dz; 

		e2->x+=e2->dx;
		e2->z+=e2->dz;

		buf_line+=gfxScrWidth;
		zbuf_line+=gfxScrWidth;
		h--;
	}
}




/********************************************************************
			DRAW FLAT TRIANGLE (GOURAUD)
*********************************************************************/

int Tri_EdgeFlatGou(TriEdge *e,TriVert *v1,TriVert *v2);
void Tri_BlitFlatGou(TriEdge *e1,TriEdge *e2,int y,int h);



void Tri_Flat_Gouraud(TriVert *v1,TriVert *v2,TriVert *v3)
{
	TriVert *p1,*p2,*p3;
	TriEdge eshort,elong;
	int y1,y2,y3,y,h;

	y1=v1->y; y2=v2->y; y3=v3->y;

	if (y1<y2) {
		if (y1<y3) {
			p1=v1;
			if (y2<y3) {
				p2=v2;
				p3=v3;
			}
			else {	// y2>y3
				p2=v3;
				p3=v2;
			}
		}
		else {	// y3<y1
			p1=v3;
			p2=v1;
			p3=v2;
		}
	}
	else { 	// y1>y2
		if (y2<y3) {
			p1=v2;
			if (y3<y1) {
				p2=v3;
				p3=v1;
			}
			else {	// y3>y1
				p2=v1;
				p3=v3;
			}
		}
		else {	// y2>y3
			p1=v3;
			p2=v2;
			p3=v1;
		}
	}

// Long side
	if (Tri_EdgeFlatGou(&elong,p1,p3)<=0) return;

// Start y
	y=max(p1->y,tri_clip_y1);

// Top short side
	h=Tri_EdgeFlatGou(&eshort,p1,p2);
	if (h>0) {
		if (eshort.dx<elong.dx)
			Tri_BlitFlatGou(&eshort,&elong,y,h);
		else
			Tri_BlitFlatGou(&elong,&eshort,y,h);
		y+=h;
	}
// Bot short side
	h=Tri_EdgeFlatGou(&eshort,p2,p3);
	if (h>0) {
		if (eshort.x<elong.x)
			Tri_BlitFlatGou(&eshort,&elong,y,h);
		else
			Tri_BlitFlatGou(&elong,&eshort,y,h);
	}
}


int Tri_EdgeFlatGou(TriEdge *e,TriVert *v1,TriVert *v2)
{
	int dy, h, adv_y=0;

	dy=h=v2->y-v1->y;

	if (v1->y<tri_clip_y1) {
		if (v2->y<=tri_clip_y1) return 0;
		adv_y=tri_clip_y1-v1->y;
		h-=adv_y;
	}
	if (v2->y>tri_clip_y2) {
		if (v1->y>=tri_clip_y2) return 0;
		h-=v2->y-tri_clip_y2;
	}

	dy=Inverse(dy);
	e->x=v1->x<<16;
	e->s=v1->s;

	e->dx=(v2->x-v1->x)*dy;
	e->ds=FixMul(v2->s-v1->s,dy);


	if (adv_y) {
		e->x+=e->dx*adv_y;
		e->s+=e->ds*adv_y;
	}

	return h;
}

void Tri_BlitFlatGou(TriEdge *e1,TriEdge *e2,int y,int h)
{
	byte *buf_line, *buf;
	int s, ds;
	int x1, x2, dx, count, x_adv;

	buf_line=gfxBuffer+y*gfxScrWidth;

	while(h>0) {

		x1=e1->x>>16; x2=e2->x>>16;
		dx=x2-x1;
		if (x2>tri_clip_x2) {
			x2=tri_clip_x2;
		}
		if (x1<tri_clip_x1) {
			x_adv=tri_clip_x1-x1;
			x1=tri_clip_x1;
		}
		else x_adv=0;

		if (x1>=x2) goto NEXT_SPAN;

		dx=Inverse(dx);
		ds=FixMul(e2->s-e1->s,dx);
		s=e1->s;

		if (x_adv) {
			s+=ds*x_adv;
		}

		buf=buf_line+x1;
		count=x2-x1;
		while(count>0) {
			*buf=tri_pal_table[((s>>8)&0xFF00)+tri_color];
			count--;
			buf++;
			s+=ds;
		}

	NEXT_SPAN:
		e1->x+=e1->dx; 
		e1->s+=e1->ds;

		e2->x+=e2->dx;
		e2->s+=e2->ds;

		buf_line+=gfxScrWidth;
		h--;
	}
}






/********************************************************************
		DRAW FLAT TRIANGLE (ZBUFFER, GOURAUD)
*********************************************************************/

int Tri_EdgeFlatGou_z(TriEdge *e,TriVert *v1,TriVert *v2);
void Tri_BlitFlatGou_zbuf(TriEdge *e1,TriEdge *e2,int y,int h);


void Tri_Flat_Gouraud_zbuf(TriVert *v1,TriVert *v2,TriVert *v3)
{
	TriVert *p1,*p2,*p3;
	TriEdge eshort,elong;
	int y1,y2,y3,y,h;

	y1=v1->y; y2=v2->y; y3=v3->y;

	if (y1<y2) {
		if (y1<y3) {
			p1=v1;
			if (y2<y3) {
				p2=v2;
				p3=v3;
			}
			else {	// y2>y3
				p2=v3;
				p3=v2;
			}
		}
		else {	// y3<y1
			p1=v3;
			p2=v1;
			p3=v2;
		}
	}
	else { 	// y1>y2
		if (y2<y3) {
			p1=v2;
			if (y3<y1) {
				p2=v3;
				p3=v1;
			}
			else {	// y3>y1
				p2=v1;
				p3=v3;
			}
		}
		else {	// y2>y3
			p1=v3;
			p2=v2;
			p3=v1;
		}
	}

// Long side
	if (Tri_EdgeFlatGou_z(&elong,p1,p3)<=0) return;

// Start y
	y=max(p1->y,tri_clip_y1);

// Top short side
	h=Tri_EdgeFlatGou_z(&eshort,p1,p2);
	if (h>0) {
		if (eshort.dx<elong.dx)
			Tri_BlitFlatGou_zbuf(&eshort,&elong,y,h);
		else
			Tri_BlitFlatGou_zbuf(&elong,&eshort,y,h);
		y+=h;
	}
// Bot short side
	h=Tri_EdgeFlatGou_z(&eshort,p2,p3);
	if (h>0) {
		if (eshort.x<elong.x)
			Tri_BlitFlatGou_zbuf(&eshort,&elong,y,h);
		else
			Tri_BlitFlatGou_zbuf(&elong,&eshort,y,h);
	}
}


int Tri_EdgeFlatGou_z(TriEdge *e,TriVert *v1,TriVert *v2)
{
	int dy, h, adv_y=0;

	dy=h=v2->y-v1->y;

	if (v1->y<tri_clip_y1) {
		if (v2->y<=tri_clip_y1) return 0;
		adv_y=tri_clip_y1-v1->y;
		h-=adv_y;
	}
	if (v2->y>tri_clip_y2) {
		if (v1->y>=tri_clip_y2) return 0;
		h-=v2->y-tri_clip_y2;
	}

	dy=Inverse(dy);
	e->x=v1->x<<16;
	e->z=v1->inv_z;
	e->s=v1->s;

	e->dx=(v2->x-v1->x)*dy;
	e->ds=FixMul(v2->s-v1->s,dy);
	e->dz=FixMul(v2->inv_z-v1->inv_z,dy);


	if (adv_y) {
		e->x+=e->dx*adv_y;
		e->z+=e->dz*adv_y;
		e->s+=e->ds*adv_y;
	}

	return h;
}

void Tri_BlitFlatGou_zbuf(TriEdge *e1,TriEdge *e2,int y,int h)
{
	byte *buf_line, *buf;
	short *zbuf_line, *zbuf, z, dz;
	int s, ds;
	int x1, x2, dx, count, x_adv;

	buf_line=gfxBuffer+y*gfxScrWidth;
	zbuf_line=ZBuffer+y*gfxScrWidth;

	while(h>0) {

		x1=e1->x>>16; x2=e2->x>>16;
		dx=x2-x1;
		if (x2>tri_clip_x2) {
			x2=tri_clip_x2;
		}
		if (x1<tri_clip_x1) {
			x_adv=tri_clip_x1-x1;
			x1=tri_clip_x1;
		}
		else x_adv=0;

		if (x1>=x2) goto NEXT_SPAN;

		dx=Inverse(dx);
		dz=FixMul(e2->z-e1->z,dx);
		ds=FixMul(e2->s-e1->s,dx);
		z=e1->z;
		s=e1->s;

		if (x_adv) {
			z+=dz*x_adv;
			s+=ds*x_adv;
		}

		buf=buf_line+x1;
		zbuf=zbuf_line+x1;
		count=x2-x1;
		while(count>0) {
			if ((*zbuf)<z) {
				*zbuf=z;
				*buf=tri_pal_table[((s>>8)&0xFF00)+tri_color];
			}
			count--;
			buf++; zbuf++; z+=dz;
			s+=ds;
		}

	NEXT_SPAN:
		e1->x+=e1->dx; 
		e1->z+=e1->dz; 
		e1->s+=e1->ds;

		e2->x+=e2->dx;
		e2->z+=e2->dz;
		e2->s+=e2->ds;

		buf_line+=gfxScrWidth;
		zbuf_line+=gfxScrWidth;
		h--;
	}
}





/********************************************************************
			DRAW TEX TRIANGLE (LAMBERT, ZBUFFER)
*********************************************************************/

void Tri_BlitTexLam_zbuf(TriEdge *e1,TriEdge *e2,int y,int h,int s);



void Tri_Tex_Lambert_zbuf(TriVert *v1,TriVert *v2,TriVert *v3)
{
	TriVert *p1,*p2,*p3;
	TriEdge eshort,elong;
	int y1,y2,y3,y,h;

	y1=v1->y; y2=v2->y; y3=v3->y;

	if (y1<y2) {
		if (y1<y3) {
			p1=v1;
			if (y2<y3) {
				p2=v2;
				p3=v3;
			}
			else {	// y2>y3
				p2=v3;
				p3=v2;
			}
		}
		else {	// y3<y1
			p1=v3;
			p2=v1;
			p3=v2;
		}
	}
	else { 	// y1>y2
		if (y2<y3) {
			p1=v2;
			if (y3<y1) {
				p2=v3;
				p3=v1;
			}
			else {	// y3>y1
				p2=v1;
				p3=v3;
			}
		}
		else {	// y2>y3
			p1=v3;
			p2=v2;
			p3=v1;
		}
	}

// Long side
	if (Tri_EdgeTex_z(&elong,p1,p3)<=0) return;

// Start y
	y=max(p1->y,tri_clip_y1);

// Top short side
	h=Tri_EdgeTex_z(&eshort,p1,p2);
	if (h>0) {
		if (eshort.dx<elong.dx)
			Tri_BlitTexLam_zbuf(&eshort,&elong,y,h,v1->s);
		else
			Tri_BlitTexLam_zbuf(&elong,&eshort,y,h,v1->s);
		y+=h;
	}
// Bot short side
	h=Tri_EdgeTex_z(&eshort,p2,p3);
	if (h>0) {
		if (eshort.x<elong.x)
			Tri_BlitTexLam_zbuf(&eshort,&elong,y,h,v1->s);
		else
			Tri_BlitTexLam_zbuf(&elong,&eshort,y,h,v1->s);
	}
}


void Tri_BlitTexLam_zbuf(TriEdge *e1,TriEdge *e2,int y,int h,int s)
{
	byte *buf_line, *buf, *pal_table;
	short *zbuf_line, *zbuf, z, dz;
	int u, v, du, dv;
	int x1, x2, dx, count, x_adv, pixel;

	pal_table=&tri_pal_table[(s>>8)&0xFF00];

	buf_line=gfxBuffer+y*gfxScrWidth;
	zbuf_line=ZBuffer+y*gfxScrWidth;

	while(h>0) {

		x1=e1->x>>16; x2=e2->x>>16;
		dx=x2-x1;
		if (x2>tri_clip_x2) {
			x2=tri_clip_x2;
		}
		if (x1<tri_clip_x1) {
			x_adv=tri_clip_x1-x1;
			x1=tri_clip_x1;
		}
		else x_adv=0;

		if (x1>=x2) goto NEXT_SPAN;

		dx=Inverse(dx);
		dz=FixMul(e2->z-e1->z,dx);
		du=FixMul(e2->u-e1->u,dx);
		dv=FixMul(e2->v-e1->v,dx);
		z=e1->z;
		u=e1->u;
		v=e1->v;

		if (x_adv) {
			z+=dz*x_adv;
			u+=du*x_adv;
			v+=dv*x_adv;
		}

		buf=buf_line+x1;
		zbuf=zbuf_line+x1;
		count=x2-x1;
		while(count>0) {
			if ((*zbuf)<z) {
				*zbuf=z;
				pixel=tri_texture[(v>>16)*tri_tex_width+(u>>16)];
				*buf=pal_table[pixel];
			}
			count--;
			buf++; zbuf++; z+=dz;
			u+=du; v+=dv;
		}

	NEXT_SPAN:
		e1->x+=e1->dx; 
		e1->z+=e1->dz; 
		e1->u+=e1->du;
		e1->v+=e1->dv;

		e2->x+=e2->dx;
		e2->z+=e2->dz;
		e2->u+=e2->du;
		e2->v+=e2->dv;

		buf_line+=gfxScrWidth;
		zbuf_line+=gfxScrWidth;
		h--;
	}
}




/********************************************************************
			DRAW TEX TRIANGLE (ZBUFFER)
*********************************************************************/

int Tri_EdgeTex_z(TriEdge *e,TriVert *v1,TriVert *v2);
void Tri_BlitTex_zbuf(TriEdge *e1,TriEdge *e2,int y,int h);



void Tri_Tex_zbuf(TriVert *v1,TriVert *v2,TriVert *v3)
{
	TriVert *p1,*p2,*p3;
	TriEdge eshort,elong;
	int y1,y2,y3,y,h;

	y1=v1->y; y2=v2->y; y3=v3->y;

	if (y1<y2) {
		if (y1<y3) {
			p1=v1;
			if (y2<y3) {
				p2=v2;
				p3=v3;
			}
			else {	// y2>y3
				p2=v3;
				p3=v2;
			}
		}
		else {	// y3<y1
			p1=v3;
			p2=v1;
			p3=v2;
		}
	}
	else { 	// y1>y2
		if (y2<y3) {
			p1=v2;
			if (y3<y1) {
				p2=v3;
				p3=v1;
			}
			else {	// y3>y1
				p2=v1;
				p3=v3;
			}
		}
		else {	// y2>y3
			p1=v3;
			p2=v2;
			p3=v1;
		}
	}

// Long side
	if (Tri_EdgeTex_z(&elong,p1,p3)<=0) return;

// Start y
	y=max(p1->y,tri_clip_y1);

// Top short side
	h=Tri_EdgeTex_z(&eshort,p1,p2);
	if (h>0) {
		if (eshort.dx<elong.dx)
			Tri_BlitTex_zbuf(&eshort,&elong,y,h);
		else
			Tri_BlitTex_zbuf(&elong,&eshort,y,h);
		y+=h;
	}
// Bot short side
	h=Tri_EdgeTex_z(&eshort,p2,p3);
	if (h>0) {
		if (eshort.x<elong.x)
			Tri_BlitTex_zbuf(&eshort,&elong,y,h);
		else
			Tri_BlitTex_zbuf(&elong,&eshort,y,h);
	}
}


int Tri_EdgeTex_z(TriEdge *e,TriVert *v1,TriVert *v2)
{
	int dy, h, adv_y=0;

	dy=h=v2->y-v1->y;

	if (v1->y<tri_clip_y1) {
		if (v2->y<=tri_clip_y1) return 0;
		adv_y=tri_clip_y1-v1->y;
		h-=adv_y;
	}
	if (v2->y>tri_clip_y2) {
		if (v1->y>=tri_clip_y2) return 0;
		h-=v2->y-tri_clip_y2;
	}

	dy=Inverse(dy);

	e->x=v1->x<<16;
	e->z=v1->inv_z;
	e->u=v1->u<<16;
	e->v=v1->v<<16;

	e->dx=(v2->x-v1->x)*dy;
	e->du=(v2->u-v1->u)*dy;
	e->dv=(v2->v-v1->v)*dy;
	e->dz=FixMul(v2->inv_z-v1->inv_z,dy);


	if (adv_y) {
		e->x+=e->dx*adv_y;
		e->z+=e->dz*adv_y;
		e->u+=e->du*adv_y;
		e->v+=e->dv*adv_y;
	}

	return h;
}

void Tri_BlitTex_zbuf(TriEdge *e1,TriEdge *e2,int y,int h)
{
	byte *buf_line, *buf;
	short *zbuf_line, *zbuf, z, dz;
	int u, v, du, dv;
	int x1, x2, dx, count, x_adv;

	buf_line=gfxBuffer+y*gfxScrWidth;
	zbuf_line=ZBuffer+y*gfxScrWidth;

	while(h>0) {

		x1=e1->x>>16; x2=e2->x>>16;
		dx=x2-x1;
		if (x2>tri_clip_x2) {
			x2=tri_clip_x2;
		}
		if (x1<tri_clip_x1) {
			x_adv=tri_clip_x1-x1;
			x1=tri_clip_x1;
		}
		else x_adv=0;

		if (x1>=x2) goto NEXT_SPAN;

		dx=Inverse(dx);
		dz=FixMul(e2->z-e1->z,dx);
		du=FixMul(e2->u-e1->u,dx);
		dv=FixMul(e2->v-e1->v,dx);
		z=e1->z;
		u=e1->u;
		v=e1->v;

		if (x_adv) {
			z+=dz*x_adv;
			u+=du*x_adv;
			v+=dv*x_adv;
		}

		buf=buf_line+x1;
		zbuf=zbuf_line+x1;
		count=x2-x1;
		while(count>0) {
			if ((*zbuf)<z) {
				*zbuf=z;
				*buf=tri_texture[(v>>16)*tri_tex_width+(u>>16)];
			}
			count--;
			buf++; zbuf++; z+=dz;
			u+=du; v+=dv;
		}

	NEXT_SPAN:
		e1->x+=e1->dx; 
		e1->z+=e1->dz; 
		e1->u+=e1->du;
		e1->v+=e1->dv;

		e2->x+=e2->dx;
		e2->z+=e2->dz;
		e2->u+=e2->du;
		e2->v+=e2->dv;

		buf_line+=gfxScrWidth;
		zbuf_line+=gfxScrWidth;
		h--;
	}
}



/********************************************************************
			DRAW TEX TRIANGLE (ZFILL)
*********************************************************************/
/*
void Tri_BlitTex_zfill(TriEdge *e1,TriEdge *e2,int y,int h);


void Tri_Tex_zfill(TriVert *v1,TriVert *v2,TriVert *v3)
{
	TriVert *p1,*p2,*p3;
	TriEdge eshort,elong;
	int y1,y2,y3,y,h;

	y1=v1->y; y2=v2->y; y3=v3->y;

	if (y1<y2) {
		if (y1<y3) {
			p1=v1;
			if (y2<y3) {
				p2=v2;
				p3=v3;
			}
			else {	// y2>y3
				p2=v3;
				p3=v2;
			}
		}
		else {	// y3<y1
			p1=v3;
			p2=v1;
			p3=v2;
		}
	}
	else { 	// y1>y2
		if (y2<y3) {
			p1=v2;
			if (y3<y1) {
				p2=v3;
				p3=v1;
			}
			else {	// y3>y1
				p2=v1;
				p3=v3;
			}
		}
		else {	// y2>y3
			p1=v3;
			p2=v2;
			p3=v1;
		}
	}

// Long side
	if (Tri_EdgeTex_z(&elong,p1,p3)<=0) return;

// Start y
	y=max(p1->y,tri_clip_y1);

// Top short side
	h=Tri_EdgeTex_z(&eshort,p1,p2);
	if (h>0) {
		if (eshort.dx<elong.dx)
			Tri_BlitTex_zfill(&eshort,&elong,y,h);
		else
			Tri_BlitTex_zfill(&elong,&eshort,y,h);
		y+=h;
	}
// Bot short side
	h=Tri_EdgeTex_z(&eshort,p2,p3);
	if (h>0) {
		if (eshort.x<elong.x)
			Tri_BlitTex_zfill(&eshort,&elong,y,h);
		else
			Tri_BlitTex_zfill(&elong,&eshort,y,h);
	}
}


void Tri_BlitTex_zfill(TriEdge *e1,TriEdge *e2,int y,int h)
{
	byte *buf_line, *buf;
	short *zbuf_line, *zbuf, z, dz;
	int u, v, du, dv;
	int x1, x2, dx, count, x_adv;

	buf_line=gfxBuffer+y*gfxScrWidth;
	zbuf_line=ZBuffer+y*gfxScrWidth;

	while(h>0) {

		x1=e1->x>>16; x2=e2->x>>16;
		dx=x2-x1;
		if (x2>tri_clip_x2) {
			x2=tri_clip_x2;
		}
		if (x1<tri_clip_x1) {
			x_adv=tri_clip_x1-x1;
			x1=tri_clip_x1;
		}
		else x_adv=0;

		if (x1>=x2) goto NEXT_SPAN;

		dx=Inverse(dx);
		dz=FixMul(e2->z-e1->z,dx);
		du=FixMul(e2->u-e1->u,dx);
		dv=FixMul(e2->v-e1->v,dx);
		z=e1->z;
		u=e1->u;
		v=e1->v;

		if (x_adv) {
			z+=dz*x_adv;
			u+=du*x_adv;
			v+=dv*x_adv;
		}

		zbuf=zbuf_line+x1;
		blit_fill_z_new(zbuf,x2-x1,z,dz);

		count=x2-x1;
		buf=buf_line+x1;
		while(count>0) {
			*buf=tri_texture[(v>>16)*tri_tex_width+(u>>16)];
			count--;
			buf++;
			u+=du; v+=dv;
		}

	NEXT_SPAN:
		e1->x+=e1->dx; 
		e1->z+=e1->dz; 
		e1->u+=e1->du;
		e1->v+=e1->dv;

		e2->x+=e2->dx;
		e2->z+=e2->dz;
		e2->u+=e2->du;
		e2->v+=e2->dv;

		buf_line+=gfxScrWidth;
		zbuf_line+=gfxScrWidth;
		h--;
	}
}
*/


/********************************************************************
		DRAW TEX TRIANGLE (GOURAUD)
*********************************************************************/

int Tri_EdgeTexGou(TriEdge *e,TriVert *v1,TriVert *v2);
void Tri_BlitTexGou(TriEdge *e1,TriEdge *e2,int y,int h);


void Tri_Tex_Gouraud(TriVert *v1,TriVert *v2,TriVert *v3)
{
	TriVert *p1,*p2,*p3;
	TriEdge eshort,elong;
	int y1,y2,y3,y,h;

	y1=v1->y; y2=v2->y; y3=v3->y;

	if (y1<y2) {
		if (y1<y3) {
			p1=v1;
			if (y2<y3) {
				p2=v2;
				p3=v3;
			}
			else {	// y2>y3
				p2=v3;
				p3=v2;
			}
		}
		else {	// y3<y1
			p1=v3;
			p2=v1;
			p3=v2;
		}
	}
	else { 	// y1>y2
		if (y2<y3) {
			p1=v2;
			if (y3<y1) {
				p2=v3;
				p3=v1;
			}
			else {	// y3>y1
				p2=v1;
				p3=v3;
			}
		}
		else {	// y2>y3
			p1=v3;
			p2=v2;
			p3=v1;
		}
	}

// Long side
	if (Tri_EdgeTexGou(&elong,p1,p3)<=0) return;

// Start y
	y=max(p1->y,tri_clip_y1);

// Top short side
	h=Tri_EdgeTexGou(&eshort,p1,p2);
	if (h>0) {
		if (eshort.dx<elong.dx)
			Tri_BlitTexGou(&eshort,&elong,y,h);
		else
			Tri_BlitTexGou(&elong,&eshort,y,h);
		y+=h;
	}
// Bot short side
	h=Tri_EdgeTexGou(&eshort,p2,p3);
	if (h>0) {
		if (eshort.x<elong.x)
			Tri_BlitTexGou(&eshort,&elong,y,h);
		else
			Tri_BlitTexGou(&elong,&eshort,y,h);
	}
}


int Tri_EdgeTexGou(TriEdge *e,TriVert *v1,TriVert *v2)
{
	int dy, h, adv_y=0;

	dy=h=v2->y-v1->y;

	if (v1->y<tri_clip_y1) {
		if (v2->y<=tri_clip_y1) return 0;
		adv_y=tri_clip_y1-v1->y;
		h-=adv_y;
	}
	if (v2->y>tri_clip_y2) {
		if (v1->y>=tri_clip_y2) return 0;
		h-=v2->y-tri_clip_y2;
	}

	dy=Inverse(dy);
	e->x=v1->x<<16;
	e->u=((int)v1->u)<<16;
	e->v=((int)v1->v)<<16;
	e->s=v1->s;

	e->dx=(v2->x-v1->x)*dy;
	e->du=((int)(v2->u-v1->u))*dy;
	e->dv=((int)(v2->v-v1->v))*dy;
	e->ds=FixMul(v2->s-v1->s,dy);


	if (adv_y) {
		e->x+=e->dx*adv_y;
		e->u+=e->du*adv_y;
		e->v+=e->dv*adv_y;
		e->s+=e->ds*adv_y;
	}

	return h;
}

void Tri_BlitTexGou(TriEdge *e1,TriEdge *e2,int y,int h)
{
	byte *buf_line, *buf;
	int u, v, s, du, dv, ds;
	int x1, x2, dx, count, x_adv;
	int pixel;

	buf_line=gfxBuffer+y*gfxScrWidth;

	while(h>0) {

		x1=e1->x>>16; x2=e2->x>>16;
		dx=x2-x1;
		if (x2>tri_clip_x2) {
			x2=tri_clip_x2;
		}
		if (x1<tri_clip_x1) {
			x_adv=tri_clip_x1-x1;
			x1=tri_clip_x1;
		}
		else x_adv=0;

		if (x1>=x2) goto NEXT_SPAN;

		dx=Inverse(dx);
		du=FixMul(e2->u-e1->u,dx);
		dv=FixMul(e2->v-e1->v,dx);
		ds=FixMul(e2->s-e1->s,dx);
		u=e1->u;
		v=e1->v;
		s=e1->s;

		if (x_adv) {
			u+=du*x_adv;
			v+=dv*x_adv;
			s+=ds*x_adv;
		}

		buf=buf_line+x1;
		count=x2-x1;
		while(count>0) {
			pixel=tri_texture[(v>>16)*tri_tex_width+(u>>16)];
			*buf=tri_pal_table[((s>>8)&0xFF00)+pixel];
			count--;
			buf++;
			u+=du; v+=dv; s+=ds;
		}

	NEXT_SPAN:
		e1->x+=e1->dx; 
		e1->u+=e1->du;
		e1->v+=e1->dv;
		e1->s+=e1->ds;

		e2->x+=e2->dx;
		e2->u+=e2->du;
		e2->v+=e2->dv;
		e2->s+=e2->ds;

		buf_line+=gfxScrWidth;
		h--;
	}
}





/********************************************************************
		DRAW TEX TRIANGLE (ZBUFFER, GOURAUD)
*********************************************************************/

int Tri_EdgeTexGou_z(TriEdge *e,TriVert *v1,TriVert *v2);
void Tri_BlitTexGou_zbuf(TriEdge *e1,TriEdge *e2,int y,int h);


void Tri_Tex_Gouraud_zbuf(TriVert *v1,TriVert *v2,TriVert *v3)
{
	TriVert *p1,*p2,*p3;
	TriEdge eshort,elong;
	int y1,y2,y3,y,h;

	y1=v1->y; y2=v2->y; y3=v3->y;

	if (y1<y2) {
		if (y1<y3) {
			p1=v1;
			if (y2<y3) {
				p2=v2;
				p3=v3;
			}
			else {	// y2>y3
				p2=v3;
				p3=v2;
			}
		}
		else {	// y3<y1
			p1=v3;
			p2=v1;
			p3=v2;
		}
	}
	else { 	// y1>y2
		if (y2<y3) {
			p1=v2;
			if (y3<y1) {
				p2=v3;
				p3=v1;
			}
			else {	// y3>y1
				p2=v1;
				p3=v3;
			}
		}
		else {	// y2>y3
			p1=v3;
			p2=v2;
			p3=v1;
		}
	}

// Long side
	if (Tri_EdgeTexGou_z(&elong,p1,p3)<=0) return;

// Start y
	y=max(p1->y,tri_clip_y1);

// Top short side
	h=Tri_EdgeTexGou_z(&eshort,p1,p2);
	if (h>0) {
		if (eshort.dx<elong.dx)
			Tri_BlitTexGou_zbuf(&eshort,&elong,y,h);
		else
			Tri_BlitTexGou_zbuf(&elong,&eshort,y,h);
		y+=h;
	}
// Bot short side
	h=Tri_EdgeTexGou_z(&eshort,p2,p3);
	if (h>0) {
		if (eshort.x<elong.x)
			Tri_BlitTexGou_zbuf(&eshort,&elong,y,h);
		else
			Tri_BlitTexGou_zbuf(&elong,&eshort,y,h);
	}
}


int Tri_EdgeTexGou_z(TriEdge *e,TriVert *v1,TriVert *v2)
{
	int dy, h, adv_y=0;

	dy=h=v2->y-v1->y;

	if (v1->y<tri_clip_y1) {
		if (v2->y<=tri_clip_y1) return 0;
		adv_y=tri_clip_y1-v1->y;
		h-=adv_y;
	}
	if (v2->y>tri_clip_y2) {
		if (v1->y>=tri_clip_y2) return 0;
		h-=v2->y-tri_clip_y2;
	}

	dy=Inverse(dy);
	e->x=v1->x<<16;
	e->z=v1->inv_z;
	e->u=((int)v1->u)<<16;
	e->v=((int)v1->v)<<16;
	e->s=v1->s;

	e->dx=(v2->x-v1->x)*dy;
	e->du=((int)(v2->u-v1->u))*dy;
	e->dv=((int)(v2->v-v1->v))*dy;
	e->ds=FixMul(v2->s-v1->s,dy);
	e->dz=FixMul(v2->inv_z-v1->inv_z,dy);


	if (adv_y) {
		e->x+=e->dx*adv_y;
		e->z+=e->dz*adv_y;
		e->u+=e->du*adv_y;
		e->v+=e->dv*adv_y;
		e->s+=e->ds*adv_y;
	}

	return h;
}

void Tri_BlitTexGou_zbuf(TriEdge *e1,TriEdge *e2,int y,int h)
{
	byte *buf_line, *buf;
	short *zbuf_line, *zbuf, z, dz;
	int u, v, s, du, dv, ds;
	int x1, x2, dx, count, x_adv;
	int pixel;

	buf_line=gfxBuffer+y*gfxScrWidth;
	zbuf_line=ZBuffer+y*gfxScrWidth;

	while(h>0) {

		x1=e1->x>>16; x2=e2->x>>16;
		dx=x2-x1;
		if (x2>tri_clip_x2) {
			x2=tri_clip_x2;
		}
		if (x1<tri_clip_x1) {
			x_adv=tri_clip_x1-x1;
			x1=tri_clip_x1;
		}
		else x_adv=0;

		if (x1>=x2) goto NEXT_SPAN;

		dx=Inverse(dx);
		dz=FixMul(e2->z-e1->z,dx);
		du=FixMul(e2->u-e1->u,dx);
		dv=FixMul(e2->v-e1->v,dx);
		ds=FixMul(e2->s-e1->s,dx);
		z=e1->z;
		u=e1->u;
		v=e1->v;
		s=e1->s;

		if (x_adv) {
			z+=dz*x_adv;
			u+=du*x_adv;
			v+=dv*x_adv;
			s+=ds*x_adv;
		}

		buf=buf_line+x1;
		zbuf=zbuf_line+x1;
		count=x2-x1;
		while(count>0) {
			if ((*zbuf)<z) {
				*zbuf=z;
				pixel=tri_texture[(v>>16)*tri_tex_width+(u>>16)];
				*buf=tri_pal_table[((s>>8)&0xFF00)+pixel];
			}
			count--;
			buf++; zbuf++; z+=dz;
			u+=du; v+=dv; s+=ds;
		}

	NEXT_SPAN:
		e1->x+=e1->dx; 
		e1->z+=e1->dz; 
		e1->u+=e1->du;
		e1->v+=e1->dv;
		e1->s+=e1->ds;

		e2->x+=e2->dx;
		e2->z+=e2->dz;
		e2->u+=e2->du;
		e2->v+=e2->dv;
		e2->s+=e2->ds;

		buf_line+=gfxScrWidth;
		zbuf_line+=gfxScrWidth;
		h--;
	}
}




/********************************************************************
		DRAW TEX TRIANGLE (ZFILL, GOURAUD)
*********************************************************************/

/*
void Tri_BlitTexGou_zfill(TriEdge *e1,TriEdge *e2,int y,int h);



void Tri_Tex_Gouraud_zfill(TriVert *v1,TriVert *v2,TriVert *v3)
{
	TriVert *p1,*p2,*p3;
	TriEdge eshort,elong;
	int y1,y2,y3,y,h;

	y1=v1->y; y2=v2->y; y3=v3->y;

	if (y1<y2) {
		if (y1<y3) {
			p1=v1;
			if (y2<y3) {
				p2=v2;
				p3=v3;
			}
			else {	// y2>y3
				p2=v3;
				p3=v2;
			}
		}
		else {	// y3<y1
			p1=v3;
			p2=v1;
			p3=v2;
		}
	}
	else { 	// y1>y2
		if (y2<y3) {
			p1=v2;
			if (y3<y1) {
				p2=v3;
				p3=v1;
			}
			else {	// y3>y1
				p2=v1;
				p3=v3;
			}
		}
		else {	// y2>y3
			p1=v3;
			p2=v2;
			p3=v1;
		}
	}

// Long side
	if (Tri_EdgeTexGou_z(&elong,p1,p3)<=0) return;

// Start y
	y=max(p1->y,tri_clip_y1);

// Top short side
	h=Tri_EdgeTexGou_z(&eshort,p1,p2);
	if (h>0) {
		if (eshort.dx<elong.dx)
			Tri_BlitTexGou_zfill(&eshort,&elong,y,h);
		else
			Tri_BlitTexGou_zfill(&elong,&eshort,y,h);
		y+=h;
	}
// Bot short side
	h=Tri_EdgeTexGou_z(&eshort,p2,p3);
	if (h>0) {
		if (eshort.x<elong.x)
			Tri_BlitTexGou_zfill(&eshort,&elong,y,h);
		else
			Tri_BlitTexGou_zfill(&elong,&eshort,y,h);
	}
}


void Tri_BlitTexGou_zfill(TriEdge *e1,TriEdge *e2,int y,int h)
{
	byte *buf_line, *buf;
 	short *zbuf_line, *zbuf, z, dz;
	int u, v, s, du, dv, ds;
	int x1, x2, dx, count, x_adv;
	int pixel;

	buf_line=gfxBuffer+y*gfxScrWidth;
	zbuf_line=ZBuffer+y*gfxScrWidth;

	while(h>0) {

		x1=e1->x>>16; x2=e2->x>>16;
		dx=x2-x1;
		if (x2>tri_clip_x2) {
			x2=tri_clip_x2;
		}
		if (x1<tri_clip_x1) {
			x_adv=tri_clip_x1-x1;
			x1=tri_clip_x1;
		}
		else x_adv=0;

		if (x1>=x2) goto NEXT_SPAN;


		dx=Inverse(dx);
		dz=FixMul(e2->z-e1->z,dx);
		du=FixMul(e2->u-e1->u,dx);
		dv=FixMul(e2->v-e1->v,dx);
		ds=FixMul(e2->s-e1->s,dx);
		z=e1->z;
		u=e1->u;
		v=e1->v;
		s=e1->s;

		if (x_adv) {
			z+=dz*x_adv;
			u+=du*x_adv;
			v+=dv*x_adv;
			s+=ds*x_adv;
		}

		zbuf=zbuf_line+x1;
		blit_fill_z_new(zbuf,x2-x1,z,dz);

		buf=buf_line+x1;
		count=x2-x1;
		while(count>0) {
			pixel=tri_texture[(v>>16)*tri_tex_width+(u>>16)];
			*buf=tri_pal_table[((s>>8)&0xFF00)+pixel];
			count--;
			buf++;
			u+=du; v+=dv; s+=ds;
		}

	NEXT_SPAN:
		e1->x+=e1->dx; 
		e1->z+=e1->dz; 
		e1->u+=e1->du;
		e1->v+=e1->dv;
		e1->s+=e1->ds;

		e2->x+=e2->dx;
		e2->z+=e2->dz;
		e2->u+=e2->du;
		e2->v+=e2->dv;
		e2->s+=e2->ds;

		buf_line+=gfxScrWidth;
		zbuf_line+=gfxScrWidth;
		h--;
	}
}
*/




/*
 Fill z buffer
*/
/*
void blit_fill_z_new(short *zbuf,int count,int z,int dz);
#pragma aux blit_fill_z_new =		\
"	test esi, 3"			\
"	jz first_not_odd"		\
"	mov [esi], ax"			\
"	dec ecx"			\
"	add esi, 2"			\
"	add eax, edx"			\
"first_not_odd:"			\
"	mov edi, ecx"			\
"	shr ecx, 1"			\
"	or ecx, ecx"			\
"	jz last_pixel_z"		\
"	mov ebx, eax"			\
"	add eax, edx"			\
"	shl eax, 16"			\
"	add eax, ebx"			\
"	mov ebx, edx"			\
"	shl dx, 1"			\
"	shl ebx, 17"			\
"	add edx, ebx"			\
"fillz_loop:"				\
"	mov [esi], eax"			\
"	add esi, 4"			\
"	add eax, edx"			\
"	dec ecx"			\
"	jnz fillz_loop"			\
"last_pixel_z:"				\
"	test edi, 1"			\
"	jz end_fillz"			\
"	mov [esi], ax"			\
"end_fillz:"				\
	parm [esi] [ecx] [eax] [edx]	\
	modify [ebx edi];
*/


