
#include "engine.h"
#include "progs\progs.h"

void ProcessPlayerAction(int player_idx);


int  Server_Init(void)
{
	NumPlayers=0;
	return 1;
}

void Server_Shut(void)
{
	FreeWorld();
	Server_Disconnect();
}


/*
 Initialize a new world.
 Returns: non-zero on success
*/
int  Server_NewWorld(char *name)
{
	int i;

// Init the world
	if (!InitWorld(name)) return 0;

// Notify all the players about new world
	for(i=0;i<NumPlayers;i++) {
		if (i==LocalPlayer) continue;
		Server_SendNewWorld(i);
	}

// Add all world entities
	Entity_NewWorld(&WorldBSP);

	return 1;
}


/*
 Update world physics, etc
*/
void Server_UpdateWorld(void)
{
	int i, status;

// Process players' actions
	for(i=0;i<NumPlayers;i++) {
		ProcessPlayerAction(i);
	}

// Update all objects
	for(i=0;i<NumObjects;i++) {
		status=Prog_Event(i,PE_UPDATE,0,0);
		if (status==1) {	// Remove object
			World_RemoveObject(i);
			i--;
		}
	}
}

void ProcessPlayerAction(int player_idx)
{
	Player *p=&Players[player_idx];


	if (p->action&ACTION_ATTACK) {
		if (p->flags&PLAYER_FLAG_DEAD) {

		// !!! body_yaw !!!

			p->flags=0;
			p->body_roll=0;
			p->head_yaw=p->head_pitch=p->head_roll=0;
			p->cam_height=0;
			p->camobj_idx=-1;
			p->obj_idx=World_AddObject(PROG_PLAYER,NULL);
			Entity_PlayerStartCoords(&WorldBSP,Objects[p->obj_idx].pob.c);
			World_ObjectMoved(p->obj_idx);
			return;
		}
		else
			Prog_Event(p->obj_idx,PE_ATTACK,0,0);
	}

	if (p->flags&PLAYER_FLAG_DEAD) return;

	if (p->action&ACTION_FORWARD)
		ApplyAcceleration(p->obj_idx,1,+1);
	if (p->action&ACTION_BACKWARD)
		ApplyAcceleration(p->obj_idx,1,-1);

	if (p->action&ACTION_MOVE_LEFT)
		ApplyAcceleration(p->obj_idx,0,-1);
	if (p->action&ACTION_MOVE_RIGHT)
		ApplyAcceleration(p->obj_idx,0,+1);

	if (p->action&ACTION_TURN_LEFT)
		ApplyAcceleration(p->obj_idx,3,+1);
	if (p->action&ACTION_TURN_RIGHT)
		ApplyAcceleration(p->obj_idx,3,-1);

	if (p->action&ACTION_JUMP)
		ApplyJump(p->obj_idx);

	if (player_idx==LocalPlayer) p->action=0;
}





static VisInfo vis;

int Server_MakeFrame(int player_idx, ClientFrame *cl)
{
	Player *p=&Players[player_idx];
	Object *cam=NULL, *src;
	POB *dest;
	BSP *bsp=&WorldBSP;
	int i,type, cam_idx, vol, pan, c;
	float coords[3], dist;


// Setup the camera/player stuff
	cam_idx=p->camobj_idx;
	if (cam_idx<0) {
		p->flags&=~PLAYER_FLAG_EXTCAMERA;
		cam_idx=p->obj_idx;
	}
	else {
		p->flags|=PLAYER_FLAG_EXTCAMERA;
	}

// ??? No object for the player ???
	if (cam_idx<0) return 0;

// Setup BSP (not really usefull for remote clients)
	cl->MainBSP=&WorldBSP;

// Fill out ClientFrame
	cam=&Objects[cam_idx];
	FVec_Copy(cam->pob.c,cl->cam_pos);
	cl->cam_angle=cam->pob.angle;

	cl->body_roll=p->body_roll;
	cl->cam_height=p->cam_height;
	cl->player_flags=p->flags;

// Clear flags that are send once
	p->flags&=~PLAYER_FLAG_START_ATTACK;

	if (player_idx!=LocalPlayer) {
		cl->head_yaw=p->head_yaw;
		cl->head_pitch=p->head_pitch;
		cl->head_roll=p->head_roll;
	}


// Compute visible BSP leaves in here
	BSP_MakeVis(bsp,cam->pob.c,&vis,1);


// Find visible objects
	cl->NumVisObjects=0;
	src=Objects; dest=cl->VisObjects;
	for(i=0;i<NumObjects;i++,src++) {
		c=WorldBSP.Leaves[src->leaf_idx].cluster;
		if (i==cam_idx||c<0) continue;
		type=Prog_GetFlags(i);
		if (type&PF_TYPE_NONE) continue;
		if (vis.VisClusters[c>>3]&(1<<(c&7))) {
			*dest=src->pob;
			dest++;
			cl->NumVisObjects++;
			if (cl->NumVisObjects>=MAX_VIS_OBJECTS) break;
		}
	}

	return 1;
}


// Temporary client storage for remote clients
static ClientFrame tempClient;

void Server_MakeFrames(ClientFrame *local) 
{
	Player *p;
	int i;

	for(i=0;i<NumPlayers;i++) {
		p=&Players[i];
		if (i==LocalPlayer)
			Server_MakeFrame(i,local);
		else {
			if (Server_MakeFrame(i,&tempClient))
				Server_SendUpdateToClient(i,&tempClient);
		}
	}

	for(i=0;i<NumObjects;i++)
		Objects[i].pob.flags=0;
}



void Server_ProcessLocalInput(ClientFrame *local)
{
	Player *p=&Players[LocalPlayer];
	Object *obj=&Objects[p->obj_idx];

	if (p->obj_idx>=0)
		obj->pob.angle+=-p->body_yaw+local->body_yaw;
	p->body_yaw=local->body_yaw;
	p->head_yaw=local->head_yaw;
	p->head_pitch=local->head_pitch;

	p->action=local->player_action;
	local->player_action=0;
}
